from astropy.io import fits
import os
from datetime import timedelta
from datetime import datetime
from glob import glob
import numpy as np
from astropy.time import Time
from suncasa.utils import fitsutils as fu

imgfitsdir = '/data1/eovsa/fits/synoptic/'
# imgfitsbkdir = '/data1/workdir/synoptic_bk/'

# imgfitsdir = '/data1/workdir/synoptic_bk/'
imgfitsbkdir = '/data1/workdir/synoptic_newbk/'


#
# def write_compress_image_fits(fname, data, header, mask=None, **kwargs):
#     """
#     Take a data header pair and write a compressed FITS file.
#
#     Parameters
#     ----------
#     fname : `str`
#         File name, with extension.
#     data : `numpy.ndarray`
#         n-dimensional data array.
#     header : `dict`
#         A header dictionary.
#     compression_type: `str`, optional
#         Compression algorithm: one of 'RICE_1', 'RICE_ONE', 'PLIO_1', 'GZIP_1', 'GZIP_2', 'HCOMPRESS_1'
#     hcomp_scale: `float`, optional
#         HCOMPRESS scale parameter
#     """
#     if kwargs is {}:
#         kwargs.update({'compression_type': 'RICE_1', 'quantize_level': 4.0})
#     if isinstance(fname, str):
#         fname = os.path.expanduser(fname)
#
#     hdunew = fits.CompImageHDU(data=data, header=header, **kwargs)
#     if mask is None:
#         hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew])
#     else:
#         hdumask = fits.CompImageHDU(data=mask.astype(np.uint8), **kwargs)
#         hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, hdumask])
#     hdulnew.writeto(fname, output_verify='fix')

def rewriteImageFits(datestr, verbose=False, writejp2=False, overwritejp2=False, overwritefits=False):
    dateobj = datetime.strptime(datestr, "%Y-%m-%d")
    datestrdir = dateobj.strftime("%Y/%m/%d/")
    imgindir = imgfitsdir + datestrdir
    imgbkdir = imgfitsbkdir + datestrdir
    if not os.path.exists(imgbkdir):
        os.makedirs(imgbkdir)

    if verbose: print('Processing EOVSA image fits files for date {}'.format(dateobj.strftime('%Y-%m-%d')))
    files = glob(os.path.join(imgindir, '*.tb.*fits'))
    files = sorted(files)
    for fl in files:
        if not os.path.exists(fl): continue
        hdul = fits.open(fl)
        if len(hdul) > 1:
            hdul.close()
            continue
        else:
            hdul.close()
        filein = os.path.join(imgbkdir, os.path.basename(fl))
        if not os.path.exists(filein):
            os.system('mv {} {}'.format(fl, filein))
        hdul = fits.open(filein)
        for hdu in hdul:
            if hdu.header['NAXIS'] == 0:
                continue
            else:
                break
        data = np.squeeze(hdu.data).copy()
        # if verbose: print('Processing {}'.format(fl))
        if overwritefits:
            if os.path.exists(fl):
                os.system('rm -f {}'.format(fl))
        if not os.path.exists(fl):
            data[np.isnan(data)] = 0.0
            fu.write_compressed_image_fits(fl, data, hdu.header, compression_type='RICE_1', quantize_level=4.0)

        fj2name = fl.replace('.fits', '.jp2')
        if writejp2:
            if overwritejp2:
                if os.path.exists(fj2name):
                    os.system('rm -f {}'.format(fj2name))
            if not os.path.exists(fj2name):
                data = np.squeeze(hdu.data).copy()
                data[np.isnan(data)] = 0.0
                fu.write_j2000_image(fj2name, data[::-1, :], hdu.header)
    return


def main(year=None, month=None, day=None, ndays=1, overwritejp2=False, overwritefits=False):
    # tst = datetime.strptime("2017-04-01", "%Y-%m-%d")
    # ted = datetime.strptime("2019-12-31", "%Y-%m-%d")
    if year:
        ted = datetime(year, month, day)
    else:
        ted = datetime.now() - timedelta(days=2)
    tst = Time(np.fix(Time(ted).mjd) - ndays + 1, format='mjd').datetime
    print("Running pipeline_fitsutils for date from {} to {}".format(tst.strftime("%Y-%m-%d"),
                                                                     ted.strftime("%Y-%m-%d")))
    dateobs = tst
    while dateobs <= ted:
        datestr = dateobs.strftime("%Y-%m-%d")
        rewriteImageFits(datestr, verbose=True, writejp2=True, overwritejp2=overwritejp2, overwritefits=overwritefits)
        dateobs = dateobs + timedelta(days=1)


if __name__ == '__main__':
    '''
    Name: 
    eovsa_fitsutils --- pipeline for created the compressed fits and jp2 files of EOVSA daily full-disk images.

    Synopsis:
    eovsa_fitsutils.py [options]... [DATE_IN_YY_MM_DD]

    Description:
    Plot EOVSA daily full-disk images at multi frequencies of the date specified
    by DATE_IN_YY_MM_DD (or from ndays before the DATE_IN_YY_MM_DD if option --ndays/-n is provided).
    If DATE_IN_YY_MM_DD is omitted, it will be set to 2 days before now by default. 
    The are no mandatory arguments in this command.

    -c, --clearcache
            Remove temporary files

    -n, --ndays
            Processing the date spanning from DATE_IN_YY_MM_DD-ndays to DATE_IN_YY_MM_DD. Default is 30

    -o, --overwritejp2
            If True, overwrite eovsa jp2 files.
            Syntax: True, False, T, F, 1, 0

    -O, --overwritefits
            If True, overwrite eovsa fits files.
            Syntax: True, False, T, F, 1, 0                          


    Example: 
    eovsa_fitsutils.py -c True -n 2 -o True -O True 2020 06 10
    '''
    import sys
    import numpy as np
    import getopt
    from datetime import datetime, timedelta

    # import subprocess
    # shell = subprocess.check_output('echo $0', shell=True).decode().replace('\n', '').split('/')[-1]
    # print("shell " + shell + " is using")

    print(sys.argv)
    year = None
    month = None
    day = None
    ndays = 1
    clearcache = True
    opts = []
    overwritejp2 = False
    overwritefits = False
    try:
        argv = sys.argv[1:]
        opts, args = getopt.getopt(argv, "c:n:o:O:", ['clearcache=', 'ndays=', 'overwritejp2=', 'overwritefits='])
        print(opts, args)
        for opt, arg in opts:
            if opt in ['-c', '--clearcache']:
                if arg in ['True', 'T', '1']:
                    clearcache = True
                elif arg in ['False', 'F', '0']:
                    clearcache = False
                else:
                    clearcache = np.bool(arg)
            elif opt in ('-n', '--ndays'):
                ndays = np.int(arg)
            elif opt in ('-o', '--overwritejp2'):
                if arg in ['True', 'T', '1']:
                    overwritejp2 = True
                elif arg in ['False', 'F', '0']:
                    overwritejp2 = False
                else:
                    overwritejp2 = np.bool(arg)
            elif opt in ('-O', '--overwritefits'):
                if arg in ['True', 'T', '1']:
                    overwritefits = True
                elif arg in ['False', 'F', '0']:
                    overwritefits = False
                else:
                    overwritefits = np.bool(arg)
        nargs = len(args)
        if nargs == 3:
            year = np.int(args[0])
            month = np.int(args[1])
            day = np.int(args[2])
        else:
            year = None
            month = None
            day = None
    except getopt.GetoptError as err:
        print(err)
        print('Error interpreting command line argument')
        year = None
        month = None
        day = None
        ndays = 1
        clearcache = True
        opts = []
        overwritejp2 = False
        overwritefits = False

    print("Running eovsa_fitsutils for date {}-{}-{}.".format(year, month, day))
    kargs = {'ndays': ndays,
             'clearcache': clearcache,
             'overwritejp2': overwritejp2,
             'overwritefits': overwritefits}
    for k, v in kargs.items():
        print(k, v)

    main(year, month, day, ndays, overwritejp2=overwritejp2, overwritefits=overwritefits)



import os
import numpy as np
from datetime import datetime
from math import *
# import jdutil
import bisect
import pdb
from taskinit import ms, tb, qa, iatool
from astropy.time import Time
from sunpy import sun
import astropy.units as u

try:
    from astropy.io import fits as pyfits
except:
    try:
        import pyfits
    except ImportError:
        raise ImportError('Neither astropy nor pyfits exists in this CASA installation')


# from astropy.constants import R_sun, au

def read_horizons(vis):
    import urllib2
    import ssl
    if not os.path.exists(vis):
        print 'Input ms data ' + vis + ' does not exist! '
        return -1
    try:
        #ms.open(vis)
        #summary = ms.summary()
        #ms.close()
        #btime = Time(summary['BeginTime'], format='mjd')
        #etime = Time(summary['EndTime'], format='mjd')
        ## alternative way to avoid conflicts with importeovsa, if needed -- more time consuming
        tb.open(vis)
        btime = Time(tb.getcell('TIME',0)/24./3600.,format='mjd')
        etime = Time(tb.getcell('TIME', tb.nrows()-1) / 24. / 3600., format='mjd')
        tb.close()
        print "Beginning time of this scan " + btime.iso
        print "End time of this scan " + etime.iso
        try:
            context = ssl._create_unverified_context()
            f = urllib2.urlopen(
                "http://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=l&TABLE_TYPE='OBSERVER'&QUANTITIES='1,17,20'&CSV_FORMAT='YES'&ANG_FORMAT='DEG'&CAL_FORMAT='BOTH'&SOLAR_ELONG='0,180'&CENTER='-81@399'&COMMAND='10'&START_TIME='" + btime.iso.replace(
                    ' ', ',') + "'&STOP_TIME='" + etime.iso[:-4].replace(' ', ',') + "'&STEP_SIZE='1m'&SKIP_DAYLT='NO'",
                context=context)
        except:
            f = urllib2.urlopen(
                "http://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=l&TABLE_TYPE='OBSERVER'&QUANTITIES='1,17,20'&CSV_FORMAT='YES'&ANG_FORMAT='DEG'&CAL_FORMAT='BOTH'&SOLAR_ELONG='0,180'&CENTER='-81@399'&COMMAND='10'&START_TIME='" + btime.iso.replace(
                    ' ', ',') + "'&STOP_TIME='" + etime.iso[:-4].replace(' ', ',') + "'&STEP_SIZE='1m'&SKIP_DAYLT='NO'")
    except:
        print 'error in reading ms file: ' + vis + ' to obtain the ephemeris!'
        return -1
    # inputs:
    #   ephemfile:
    #       OBSERVER output from JPL Horizons for topocentric coordinates with for example
    #       target=Sun, observer=VLA=-5
    #       extra precision, quantities 1,17,20, REFRACTION
    #       routine goes through file to find $$SOE which is start of ephemeris and ends with $$EOE
    # outputs: a Python dictionary containing the following:
    #   timestr: date and time as a string
    #   time: modified Julian date
    #   ra: right ascention, in rad
    #   dec: declination, in rad
    #   rastr: ra in string
    #   decstr: dec in string
    #   p0: solar p angle, CCW with respect to the celestial north pole
    #   delta: distance from the disk center to the observer, in AU
    #   delta_dot: time derivative of delta, in the light of sight direction. Negative means it is moving toward the observer
    #
    # initialize the return dictionary
    ephem0 = dict.fromkeys(['time', 'ra', 'dec', 'delta', 'p0'])
    lines = f.readlines()
    f.close()
    nline = len(lines)
    istart = 0
    for i in range(nline):
        line = lines[i]
        if line[0:5] == '$$SOE':  # start recording
            istart = i + 1
        if line[0:5] == '$$EOE':  # end recording
            iend = i
    newlines = lines[istart:iend]
    nrec = len(newlines)
    ephem_ = []
    t = []
    ra = []
    dec = []
    p0 = []
    delta = []
    for line in newlines:
        items = line.split(',')
        # t.append({'unit':'mjd','value':Time(float(items[1]),format='jd').mjd})
        # ra.append({'unit': 'rad', 'value': np.radians(float(items[4]))})
        # dec.append({'unit': 'rad', 'value': np.radians(float(items[5]))})
        # p0.append({'unit': 'deg', 'value': float(items[6])})
        # delta.append({'unit': 'au', 'value': float(items[8])})
        t.append(Time(float(items[1]), format='jd').mjd)
        ra.append(np.radians(float(items[4])))
        dec.append(np.radians(float(items[5])))
        p0.append(float(items[6]))
        delta.append(float(items[8]))
    # convert list of dictionary to a dictionary of arrays
    ephem = {'time': t, 'ra': ra, 'dec': dec, 'p0': p0, 'delta': delta}
    return ephem


def read_msinfo(vis=None, msinfofile=None):
    # read MS information #
    msinfo = dict.fromkeys(['vis', 'scans', 'fieldids', 'btimes', 'btimestr', \
                            'inttimes', 'ras', 'decs'])
    ms.open(vis)
    scans = ms.getscansummary()
    scanids = sorted(scans.keys(), key=lambda x: int(x))
    nscanid = len(scanids)
    btimes = []
    btimestr = []
    etimes = []
    fieldids = []
    inttimes = []
    dirs = []
    ras = []
    decs = []
    for i in range(nscanid):
        btimes.append(scans[scanids[i]]['0']['BeginTime'])
        etimes.append(scans[scanids[i]]['0']['EndTime'])
        fieldid = scans[scanids[i]]['0']['FieldId']
        fieldids.append(fieldid)
        dir = ms.getfielddirmeas('PHASE_DIR', fieldid)
        dirs.append(dir)
        ras.append(dir['m0'])
        decs.append(dir['m1'])
        inttimes.append(scans[scanids[i]]['0']['IntegrationTime'])
    ms.close()
    btimestr = [qa.time(qa.quantity(btimes[i], 'd'), form='fits', prec=10)[0] \
                for i in range(nscanid)]
    msinfo['vis'] = vis
    msinfo['scans'] = scans
    msinfo['fieldids'] = fieldids
    msinfo['btimes'] = btimes
    msinfo['btimestr'] = btimestr
    msinfo['inttimes'] = inttimes
    msinfo['ras'] = ras
    msinfo['decs'] = decs
    if msinfofile:
        np.savez(msinfofile, vis=vis, scans=scans, fieldids=fieldids, \
                 btimes=btimes, btimestr=btimestr, inttimes=inttimes, \
                 ras=ras, decs=decs)
    return msinfo

def ephem_to_helio(vis=None, ephem=None, msinfo=None, reftime=None, polyfit=None, usephacenter=False):
    '''1. Take a solar ms database, read the scan and field information, find out the pointings (in RA and DEC)
       2. Compare with the ephemeris of the solar disk center (in RA and DEC)
       3. Generate VLA pointings in heliocentric coordinates
       inputs:
           msinfo: CASA MS information, output from read_msinfo
           ephem: solar ephem, output from read_horizons
           reftime: list of reference times (e.g., used for imaging)
                    CASA standard time format, either a single time (e.g., '2012/03/03/12:00:00'
                    or a time range (e.g., '2012/03/03/12:00:00~2012/03/03/13:00:00'. If the latter,
                    take the midpoint of the timerange for reference. If no date specified, take
                    the date of the first scan
           polyfit: ONLY works for MS database with only one source with continously tracking; 
                    not recommanded unless scan length is too long and want to have very high accuracy
           usephacenter: Bool -- if True, correct for the RA and DEC in the ms file based on solar empheris. 
                                 Otherwise assume the phasecenter is correctly pointed to the solar disk center
                                 (EOVSA case)
     
       return value:
           helio: a list of VLA pointing information
                   reftimestr: reference time, in FITS format string
                   reftime: reference time, in mjd format
                   ra: actual RA of phasecenter in the ms file at the reference time (interpolated)
                   dec: actual DEC of phasecenter in the ms file at the reference time (interpolated)
                   # CASA uses only RA and DEC of the closest field (e.g. in clean) #
                   ra_fld: right ascention of the CASA reference pointing direction
                   dec_fld: declination of the CASA reference pointing direction
                   raoff: RA offset of the phasecenter in the ms file to solar center
                   decoff: DEC offset of the phasecenter in the ms file to solar center
                   refx: heliocentric X offset of the phasecenter in the ms file to solar center
                   refy: heliocentric Y offset of the phasecenter in the ms file to solar center
    ######## Example #########
         msfile='sun_C_20140910T221952-222952.10s.cal.ms'
         ephemfile='horizons_sun_20140910.radecp'
         ephem=vla_prep.read_horizons(ephemfile=ephemfile)
         msinfo=vla_prep.read_msinfo(msfile=msfile)
         polyfit=0
         reftime = '22:25:20~22:25:40'
    '''
    if not vis or not os.path.exists(vis):
        raise ValueError, 'Please provide information of the MS database!'
    if not ephem:
        ephem = read_horizons(vis)
    if not msinfo:
        msinfo0 = read_msinfo(vis)
    else:
        if isinstance(msinfo, str):
            try:
                msinfo0 = np.load(msinfo)
            except:
                raise ValueError, 'The specified input msinfo file does not exist!'
        elif isinstance(msinfo, dict):
            msinfo0 = msinfo
        else:
            raise ValueError, 'msinfo should be either a numpy npz or a dictionary'
    print 'msinfo is derived from: ', msinfo0['vis']
    scans = msinfo0['scans']
    fieldids = msinfo0['fieldids']
    btimes = msinfo0['btimes']
    inttimes = msinfo0['inttimes']
    ras = msinfo0['ras']
    decs = msinfo0['decs']
    ra_rads = [ra['value'] for ra in ras]
    dec_rads = [dec['value'] for dec in decs]
    # fit 2nd order polynomial fits to the RAs and DECs #
    if polyfit:
        cra = np.polyfit(btimes, ra_rads, 2)
        cdec = np.polyfit(btimes, dec_rads, 2)

    # find out phase center infomation in ms according to the input time or timerange #
    if not reftime:
        raise ValueError, 'Please specify a reference time of the image'
    if type(reftime) == str:
        reftime = [reftime]
    if (not isinstance(reftime, list)):
        print 'input "reftime" is not a valid list. Abort...'

    nreftime = len(reftime)
    helio = []
    for reftime0 in reftime:
        helio0 = dict.fromkeys(['reftimestr', 'reftime', \
                                'ra', 'dec', 'ra_fld', 'dec_fld', \
                                'raoff', 'decoff', 'refx', 'refy', 'p0'])
        helio0['reftimestr'] = reftime0
        if '~' in reftime0:
            # if reftime0 is specified as a timerange
            [tbg0, tend0] = reftime0.split('~')
            tbg_d = qa.getvalue(qa.convert(qa.totime(tbg0), 'd'))[0]
            tend_d = qa.getvalue(qa.convert(qa.totime(tend0), 'd'))[0]
            tdur_s = (tend_d - tbg_d) * 3600. * 24.
            # if no date is specified, add up the date of the first scan
            if tend_d < 1.:
                if tend_d >= tbg_d:
                    tend_d += int(btimes[0])
                else:
                    tend_d += int(btimes[0]) + 1
            if tbg_d < 1.:
                tbg_d += int(btimes[0])
            tref_d = (tbg_d + tend_d) / 2.
        else:
            # if reftime0 is specified as a single value
            tref_d = qa.getvalue(qa.convert(qa.totime(reftime0), 'd'))
            # if no date is specified, add up the date of the first scan
            if tref_d < 1.:
                tref_d += int(btimes[0])
            tbg_d = tref_d
            # use the intergration time
            ind = bisect.bisect_left(btimes, tref_d)
            tdur_s = inttims[ind - 1]
        helio0['reftime'] = tref_d
        helio0['date-obs'] = qa.time(qa.quantity(tbg_d, 'd'), form='fits', prec=10)[0]
        helio0['exptime'] = tdur_s

        # find out phase center RA and DEC in the measurement set according to the reference time
        # if polyfit, then use the 2nd order polynomial coeffs
        ind = bisect.bisect_left(btimes, tref_d)
        if ind > 1:
            dt = tref_d - btimes[ind - 1]
            if ind < len(btimes):
                scanlen = btimes[ind] - btimes[ind - 1]
                (ra_b, ra_e) = (ras[ind - 1]['value'], ras[ind]['value'])
                (dec_b, dec_e) = (decs[ind - 1]['value'], decs[ind]['value'])
            if ind >= len(btimes):
                scanlen = btimes[ind - 1] - btimes[ind - 2]
                (ra_b, ra_e) = (ras[ind - 2]['value'], ras[ind - 1]['value'])
                (dec_b, dec_e) = (decs[ind - 2]['value'], decs[ind - 1]['value'])
        if ind == 1:  # only one scan exists (e.g., imported from AIPS)
            ra_b = ras[ind - 1]['value']
            ra_e = ra_b
            dec_b = decs[ind - 1]['value']
            dec_e = dec_b
            scanlen = 10.  # radom value
            dt = 0.
        if ind < 1:
            raise ValueError, 'Reference time does not fall into the scan list!'
        if polyfit:
            ra = cra[0] * tref_d ** 2. + cra[1] * tref_d + cra[2]
            dec = cdec[0] * tref_d ** 2. + cdec[1] * tref_d + cdec[2]
        # if not, use linearly interpolated RA and DEC at the beginning of this scan and next scan
        else:
            ra = ra_b + (ra_e - ra_b) / scanlen * dt
            dec = dec_b + (dec_e - dec_b) / scanlen * dt
        if ra < 0:
            ra += 2. * np.pi
        if ra_b < 0:
            ra_b += 2. * np.pi

        # compare with ephemeris from JPL Horizons
        time0 = ephem['time']
        ra0 = ephem['ra']
        dec0 = ephem['dec']
        p0 = ephem['p0']
        delta0 = ephem['delta']
        ind = bisect.bisect_left(time0, tref_d)
        dt0 = time0[ind] - time0[ind - 1]
        dt_ref = tref_d - time0[ind - 1]
        dra0 = ra0[ind] - ra0[ind - 1]
        ddec0 = dec0[ind] - dec0[ind - 1]
        dp0 = p0[ind] - p0[ind - 1]
        ddelta0 = delta0[ind] - delta0[ind - 1]
        ra0 = ra0[ind - 1] + dra0 / dt0 * dt_ref
        dec0 = dec0[ind - 1] + ddec0 / dt0 * dt_ref
        p0 = p0[ind - 1] + dp0 / dt0 * dt_ref
        delta0 = delta0[ind - 1] + ddelta0 / dt0 * dt_ref
        if ra0 < 0:
            ra0 += 2. * np.pi

        # RA and DEC offset in arcseconds
        decoff = degrees((dec - dec0)) * 3600.
        raoff = degrees((ra - ra0) * cos(dec)) * 3600.
        # Convert into heliocentric offsets
        prad = -radians(p0)
        refx = (-raoff) * cos(prad) - decoff * sin(prad)
        refy = (-raoff) * sin(prad) + decoff * cos(prad)
        helio0['ra'] = ra  # ra of the actual pointing
        helio0['dec'] = dec  # dec of the actual pointing
        helio0['ra_fld'] = ra_b  # ra of the field, used as the reference in e.g., clean
        helio0['dec_fld'] = dec_b  # dec of the field, used as the refenrence in e.g., clean
        helio0['raoff'] = raoff
        helio0['decoff'] = decoff
        if usephacenter:
            helio0['refx'] = refx
            helio0['refy'] = refy
        else:
            helio0['refx'] = 0.
            helio0['refy'] = 0.
        helio0['p0'] = p0
        # helio['r_sun']=np.degrees(R_sun.value/(au.value*delta0))*3600. #in arcsecs
        helio.append(helio0)
    return helio


def getbeam(imagefile=None, beamfile=None):
    ia = iatool()
    if not imagefile:
        raise ValueError, 'Please specify input images'
    bmaj = []
    bmin = []
    bpa = []
    beamunit = []
    bpaunit = []
    chans = []
    nimg = len(imagefile)
    for n in range(nimg):
        img = imagefile[n]
        if not os.path.exists(img):
            raise ValueError, 'The input image does not exist!'
        ia.open(img)
        sum = ia.summary()
        bmaj_ = []
        bmin_ = []
        bpa_ = []
        if sum.has_key('perplanebeams'):  # beam vary with frequency
            nbeams = sum['perplanebeams']['nChannels']
            beams = sum['perplanebeams']['beams']
            chans_ = [key[1:] for key in beams.keys()]
            chans_.sort(key=float)
            for chan in chans_:
                bmaj0 = beams['*' + chan]['*0']['major']['value']
                bmaj_.append(bmaj0)
                bmin0 = beams['*' + chan]['*0']['minor']['value']
                bmin_.append(bmin0)
                bpa0 = beams['*' + chan]['*0']['positionangle']['value']
                bpa_.append(bpa0)
            beamunit_ = beams['*' + chans_[0]]['*0']['major']['unit']
            bpaunit_ = beams['*' + chans_[0]]['*0']['positionangle']['unit']
        if sum.has_key('restoringbeam'):  # only one beam
            bmaj_.append(sum['restoringbeam']['major']['value'])
            bmin_.append(sum['restoringbeam']['minor']['value'])
            bpa_.append(sum['restoringbeam']['positionangle']['value'])
            beamunit_ = sum['restoringbeam']['major']['unit']
            bpaunit_ = sum['restoringbeam']['positionangle']['unit']
            nbeams = 1
            chans_ = [0]

        bmaj.append(bmaj_)
        bmin.append(bmin_)
        bpa.append(bpa_)
        beamunit.append(beamunit_)
        bpaunit.append(bpaunit_)
        chans.append(chans_)
    if beamfile:  # write beams to ascii file
        print 'Writing beam info to ascii file...'
        f = open(beamfile, 'w')
        f.write('CHANNEL No., BMAJ (' + beamunit[0] + '), BMIN (' + beamunit[0] + '), BPA (' + bpaunit[0] + ')')
        f.write("\n")
        for n in range(nimg):
            f.write('----For image: ' + imagefile[n] + '----')
            f.write('\n')
            chans_ = chans[n]
            for i in range(len(chans_)):
                f.write(str(chans_[i]) + ', ' + str(bmaj[n][i]) + ', ' + str(bmin[n][i]) + ', ' + str(bpa[n][i]))
                f.write("\n")
        f.close()
    return bmaj, bmin, bpa, beamunit, bpaunit


def imreg(vis=None, ephem=None, msinfo=None, reftime=None, imagefile=None, fitsfile=None, beamfile=None, \
          offsetfile=None, toTb=None, scl100=None, verbose=False, p_ang = False, overwrite = True, usephacenter=False):
    ia=iatool()
    if not imagefile:
        raise ValueError, 'Please specify input image'
    if not reftime:
        raise ValueError, 'Please specify reference time corresponding to the input image'
    if not fitsfile:
        fitsfile = [img + '.fits' for img in imagefile]
    if len(imagefile) != len(reftime):
        raise ValueError, 'Number of input images does not equal to number of helio coord headers!'
    if len(imagefile) != len(fitsfile):
        raise ValueError, 'Number of input images does not equal to number of output fits files!'
    nimg = len(imagefile)
    if verbose:
        print str(nimg) + ' images to process...'
    helio = ephem_to_helio(vis,ephem=ephem,msinfo=msinfo,reftime=reftime,usephacenter=usephacenter)
    for n, img in enumerate(imagefile):
        if verbose:
            print 'processing image #' + str(n)
        fitsf = fitsfile[n]
        hel = helio[n]
        if not os.path.exists(img):
            raise ValueError, 'Please specify input image'
        if os.path.exists(fitsf) and not overwrite:
            raise ValueError, 'Specified fits file already exists and overwrite is set to False. Aborting...'
        else:
            p0 = hel['p0']
            ia.open(img)
            imr = ia.rotate(pa=str(-p0) + 'deg')
            imr.tofits(fitsf, history=False, overwrite=overwrite)
            imr.close()
            sum = ia.summary()
            ia.close()
        # construct the standard fits header
        # RA and DEC of the reference pixel crpix1 and crpix2
        (imra, imdec) = (sum['refval'][0], sum['refval'][1])
        # find out the difference of the image center to the CASA phase center
        # RA and DEC difference in arcseconds
        ddec = degrees((imdec - hel['dec_fld'])) * 3600.
        dra = degrees((imra - hel['ra_fld']) * cos(hel['dec_fld'])) * 3600.
        # Convert into image heliocentric offsets
        prad = -radians(hel['p0'])
        dx = (-dra) * cos(prad) - ddec * sin(prad)
        dy = (-dra) * sin(prad) + ddec * cos(prad)
        if offsetfile:
            try:
                offset = np.load(offsetfile)
            except:
                raise ValueError, 'The specified offsetfile does not exist!'
            reftimes_d = offset['reftimes_d']
            xoffs = offset['xoffs']
            yoffs = offset['yoffs']
            timg_d = hel['reftime']
            ind = bisect.bisect_left(reftimes_d, timg_d)
            xoff = xoffs[ind - 1]
            yoff = yoffs[ind - 1]
        else:
            xoff = hel['refx']
            yoff = hel['refy']
        if verbose:
            print 'offset of image phase center to visibility phase center (arcsec): ', dx, dy
            print 'offset of visibility phase center to solar disk center (arcsec): ', xoff, yoff
        (crval1, crval2) = (xoff + dx, yoff + dy)
        # update the fits header to heliocentric coordinates

        hdu = pyfits.open(fitsf, mode='update')
        header = hdu[0].header
        (cdelt1, cdelt2) = (
            -header['cdelt1'] * 3600., header['cdelt2'] * 3600.)  # Original CDELT1, 2 are for RA and DEC in degrees
        header['cdelt1'] = cdelt1
        header['cdelt2'] = cdelt2
        header['cunit1'] = 'arcsec'
        header['cunit2'] = 'arcsec'
        header['crval1'] = crval1
        header['crval2'] = crval2
        header['ctype1'] = 'HPLN-TAN'
        header['ctype2'] = 'HPLT-TAN'
        header['date-obs'] = hel['date-obs'] #begin time of the image
        if not p_ang:
            hel['p0'] = 0
        try:
            # this works for pyfits version of CASA 4.7.0 but not CASA 4.6.0
            header.update('exptime', hel['exptime'])
            header.update('p_angle', hel['p0'])
            header.update('dsun_obs', sun.sunearth_distance(Time(hel['date-obs'])).to(u.meter).value)
            header.update('rsun_obs', sun.solar_semidiameter_angular_size(Time(hel['date-obs'])).value)
            header.update('rsun_ref', sun.constants.radius.value)
            header.update('hgln_obs', 0.)
            header.update('hglt_obs', sun.heliographic_solar_center(Time(hel['date-obs']))[1].value)
        except:
            # this works for astropy.io.fits
            header.append(('exptime', hel['exptime']))
            header.append(('p_angle', hel['p0']))
            header.append(('dsun_obs', sun.sunearth_distance(Time(hel['date-obs'])).to(u.meter).value))
            header.append(('rsun_obs', sun.solar_semidiameter_angular_size(Time(hel['date-obs'])).value))
            header.append(('rsun_ref', sun.constants.radius.value))
            header.append(('hgln_obs', 0.))
            header.append(('hglt_obs', sun.heliographic_solar_center(Time(hel['date-obs']))[1].value))

        # header.update('comment', 'Fits header updated to heliocentric coordinates by Bin Chen')

        # update intensity units, i.e. to brightness temperature?
        if toTb:
            # get restoring beam info
            (bmajs, bmins, bpas, beamunits, bpaunits) = getbeam(imagefile=imagefile, beamfile=beamfile)
            bmaj = bmajs[n]
            bmin = bmins[n]
            beamunit = beamunits[n]
            data = hdu[0].data  # remember the data order is reversed due to the FITS convension
            dim = data.ndim
            sz = data.shape
            keys = header.keys()
            values = header.values()
            # which axis is frequency?
            faxis = keys[values.index('FREQ')][-1]
            faxis_ind = dim - int(faxis)
            if header['BUNIT'].lower() == 'jy/beam':
                header['BUNIT'] = 'K'
                for i in range(sz[faxis_ind]):
                    nu = header['CRVAL' + faxis] + header['CDELT' + faxis] * \
                                                   (i + 1 - header['CRPIX' + faxis])
                    if header['CUNIT' + faxis] == 'KHz':
                        nu *= 1e3
                    if header['CUNIT' + faxis] == 'MHz':
                        nu *= 1e6
                    if header['CUNIT' + faxis] == 'GHz':
                        nu *= 1e9
                    if len(bmaj) > 1:  # multiple (per-plane) beams
                        bmajtmp = bmaj[i]
                        bmintmp = bmin[i]
                    else:  # one single beam
                        bmajtmp = bmaj[0]
                        bmintmp = bmin[0]
                    if beamunit == 'arcsec':
                        bmaj0 = np.radians(bmajtmp / 3600.)
                        bmin0 = np.radians(bmajtmp / 3600.)
                    if beamunit == 'arcmin':
                        bmaj0 = np.radians(bmajtmp / 60.)
                        bmin0 = np.radians(bmintmp / 60.)
                    if beamunit == 'deg':
                        bmaj0 = np.radians(bmajtmp)
                        bmin0 = np.radians(bmintmp)
                    if beamunit == 'rad':
                        bmaj0 = bmajtmp
                        bmin0 = bmintmp
                    beam_area = bmaj0 * bmin0 * np.pi / (4. * log(2.))
                    k_b = qa.constants('k')['value']
                    c_l = qa.constants('c')['value']
                    factor = 2. * k_b * nu ** 2 / c_l ** 2  # SI unit
                    jy_to_si = 1e-26
                    # print nu/1e9, beam_area, factor
                    factor2 = 1.
                    if scl100:
                        factor2 = 100.
                    if faxis == '3':
                        data[:, i, :, :] *= jy_to_si / beam_area / factor * factor2
                    if faxis == '4':
                        data[i, :, :, :] *= jy_to_si / beam_area / factor * factor2

        hdu.flush()
        hdu.close()



import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pylab as plt
import scipy.optimize as opt
from bokeh.models import (ColumnDataSource)
import matplotlib.cm as cm

__author__ = ["Sijie Yu"]
__email__ = "sijie.yu@njit.edu"


def get_contour_data(X, Y, Z):
    cs = plt.contour(X, Y, Z, levels=(np.arange(5, 10, 2) / 10.0 * np.nanmax(Z)).tolist(), cmap=cm.Greys_r)
    xs = []
    ys = []
    xt = []
    yt = []
    col = []
    text = []
    isolevelid = 0
    for isolevel in cs.collections:
        # isocol = isolevel.get_color()[0]
        # thecol = 3 * [None]
        theiso = '{:.0f}%'.format(cs.get_array()[isolevelid] / Z.max() * 100)
        isolevelid += 1
        # for i in xrange(3):
        # thecol[i] = int(255 * isocol[i])
        thecol = '#%02x%02x%02x' % (220, 220, 220)
        # thecol = '#03FFF9'

        for path in isolevel.get_paths():
            v = path.vertices
            x = v[:, 0]
            y = v[:, 1]
            xs.append(x.tolist())
            ys.append(y.tolist())
            xt.append(x[len(x) / 2])
            yt.append(y[len(y) / 2])
            text.append(theiso)
            col.append(thecol)

    source = ColumnDataSource(data={'xs': xs, 'ys': ys, 'line_color': col, 'xt': xt, 'yt': yt, 'text': text})
    return source


def twoD_Gaussian((x, y), amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    xo = float(xo)
    yo = float(yo)
    a = (np.cos(theta) ** 2) / (2 * sigma_x ** 2) + (np.sin(theta) ** 2) / (2 * sigma_y ** 2)
    b = -(np.sin(2 * theta)) / (4 * sigma_x ** 2) + (np.sin(2 * theta)) / (4 * sigma_y ** 2)
    c = (np.sin(theta) ** 2) / (2 * sigma_x ** 2) + (np.cos(theta) ** 2) / (2 * sigma_y ** 2)
    g = offset + amplitude * np.exp(- (a * ((x - xo) ** 2) + 2 * b * (x - xo) * (y - yo) + c * ((y - yo) ** 2)))
    return g.ravel()


# def maxfit(image):
#     # NAME:
#     #    maxfit
#     # PURPOSE:
#     #    find maximum position of an image using 2D Gaussian fit
#     # INPUTS:
#     #    image: sunpy
#
#     data = image.data
#     nx, ny = data.shape
#     dx, dy = image.scale.x.value, image.scale.y.value
#     xc, yc = image.center.x.value, image.center.y.value
#     mapx, mapy = (np.linspace(0, nx - 1, nx) - image.reference_pixel.x.value + 1 + 0.5) * dx + xc, (
#         np.linspace(0, ny - 1, ny) - image.reference_pixel.y.value + 1 + 0.5) * dy + yc
#     mapx, mapy = np.meshgrid(mapx, mapy)
#     idxmax = np.where(data == np.amax(data))
#     idxhm = np.where(data >= np.amax(data) / 2)
#     hmfw = np.sqrt(len(idxhm[0]) / np.pi) * dx
#     xmax, ymax = mapx[idxmax[0][0], idxmax[1][0]], mapy[idxmax[0][0], idxmax[1][0]]
#     theta = np.arctan(
#         abs(idxhm[1] - image.reference_pixel.y.value).sum() / abs(idxhm[0] - image.reference_pixel.x.value).sum())
#     initial_guess = (np.amax(data), xmax, ymax, hmfw, hmfw, theta, data.std())
#
#     try:
#         popt, pcov = opt.curve_fit(twoD_Gaussian, (mapx, mapy), data.ravel(), p0=initial_guess)
#     except:
#         popt = np.empty((7))
#         popt[:] = np.nan
#
#     return popt




try:
    ## Full Installation of CASA 4, 5 and 6
    from taskinit import ms, tb, qa
    from clearcal_cli import clearcal_cli as clearcal
    from split_cli import split_cli as split
except:
    ## Modular Installation of CASA 6
    from casatools import table as tbtool
    from casatools import ms as mstool
    from casatools import quanta as qatool
    from casatasks import split, clearcal

    tb = tbtool()
    ms = mstool()
    qa = qatool()

import numpy as np
from tqdm import tqdm
import os


def get_bandinfo(msfile, spw=None, returnbdinfo=False):
    '''
    get center frequencies of all spectral windows for msfile
    spw: [option] return the cfreq of spw. spw can be a a string or a list of string.
    The syntax of spw follows the standard spw Parameter in CASA
    if spw is not provided, return the cfreq of all spws in the msfile.
    return cfreqs is in GHz
    if returnbounds is True, return a dictionary including comprehensive freq information of the ms.
    '''

    ms.open(msfile)
    spwInfo = ms.getspectralwindowinfo()
    nspw = len(spwInfo.keys())
    reffreqs = []
    bdwds = []
    chanwds = []
    nchans = []
    for s in range(nspw):
        s_ = str(s)
        reffreqs.append(spwInfo[s_]['RefFreq'])
        bdwds.append(spwInfo[s_]['TotalWidth'])
        chanwds.append(spwInfo[s_]['ChanWidth'])
        nchans.append(spwInfo[s_]['NumChan'])
    reffreqs = np.array(reffreqs) / 1e9
    bdwds = np.array(bdwds) / 1e9
    chanwds = np.array(chanwds) / 1e9
    nchans = np.array(nchans)
    ms.close()
    cfreqs = reffreqs + bdwds / 2.0 - chanwds / 2.0
    bdinfo = {'bounds_all': np.hstack((reffreqs, reffreqs[-1] + bdwds[-1])), 'cfreqs_all': cfreqs,
              'bounds_all_lo': reffreqs, 'bounds_all_hi': reffreqs + bdwds}
    if spw is not None:
        freqbounds_lo_spw = []
        freqbounds_hi_spw = []
        cfreqs_spw = []
        for sp in spw:
            if ':' in sp:
                sp_, chn = sp.split(':')
                sp_ = int(sp_)
                if '~' in chn:
                    chns = chn.split('~')
                    freqbound_lo = reffreqs[sp_] + chanwds[sp_] * int(chns[0])
                    freqbound_hi = reffreqs[sp_] + chanwds[sp_] * (int(chns[1]) + 1)
                else:
                    ch = max(0, min(int(chn), nchans[sp_] - 1))
                    freqbound_lo = reffreqs[sp_] + chanwds[sp_] * ch
                    freqbound_hi = reffreqs[sp_] + chanwds[sp_] * (ch+1)
                cfreq = (freqbound_lo + freqbound_hi) / 2.0
            else:
                if '~' in sp:
                    sps = sp.split('~')
                    cfreq = np.nanmean([cfreqs[max(0, min(int(s), nspw - 1))] for s in sps])
                    s1 = max(0, min(int(sps[0]), nspw - 1))
                    s2 = max(0, min(int(sps[1]) + 1, nspw - 1))
                    freqbound_lo = bdinfo['bounds_all'][s1]
                    freqbound_hi = bdinfo['bounds_all'][s2]
                else:
                    s = max(0, min(int(sp), nspw - 1))
                    cfreq = cfreqs[s]
                    freqbound_lo = bdinfo['bounds_all'][s]
                    # freqbound_hi = freqbounds['bounds_all'][max(0, min(int(sp) + 1, nspw - 1))]
                    freqbound_hi = bdinfo['bounds_all'][s] + bdwds[s]
            freqbounds_lo_spw.append(freqbound_lo)
            freqbounds_hi_spw.append(freqbound_hi)
            cfreqs_spw.append(cfreq)
        cfreqs = np.array(cfreqs_spw)
        freqbounds_lo = np.array(freqbounds_lo_spw)
        freqbounds_hi = np.array(freqbounds_hi_spw)
        bdinfo['bounds_lo'] = freqbounds_lo
        bdinfo['bounds_hi'] = freqbounds_hi
        bdinfo['cfreqs'] = cfreqs
    if returnbdinfo:
        return bdinfo
    else:
        return cfreqs


def get_bmsize(cfreq, refbmsize=30.0, reffreq=1.6, minbmsize=4.0):
    '''
    get beamsize at frequencies definded by cfreq based on refbmsize at reffreq
    cfreq: input frequencies at GHz
    refbmsize: reference beam size in arcsec
    reffreq: reference frequency in GHz
    minbmsize: minimum beam size in arcsec
    '''
    bmsize = refbmsize * reffreq / cfreq
    bmsize[bmsize < minbmsize] = minbmsize
    return bmsize


def get_trange(msfile):
    from astropy.time import Time
    tb.open(msfile)
    tr = np.array([tb.getcell('TIME', 0), tb.getcell('TIME', tb.nrows() - 1)]) / 24. / 3600.
    tb.close()
    return Time(tr, format='mjd')


def time2filename(msfile, timerange='', spw=''):
    from astropy.time import Time
    tb.open(msfile)
    starttim = Time(tb.getcell('TIME', 0) / 24. / 3600., format='mjd')
    endtim = Time(tb.getcell('TIME', tb.nrows() - 1) / 24. / 3600., format='mjd')
    tb.close()
    datstr = starttim.iso[:10]
    ms.open(msfile)
    metadata = ms.metadata()
    observatory = metadata.observatorynames()[0]
    ms.close()
    if timerange is None or timerange == '':
        starttim1 = starttim
        endtim1 = endtim
    else:
        (tstart, tend) = timerange.split('~')
        if tstart[2] == ':':
            starttim1 = Time(datstr + 'T' + tstart)
            endtim1 = Time(datstr + 'T' + tend)
        else:
            starttim1 = Time(qa.quantity(tstart, 'd')['value'], format='mjd')
            endtim1 = Time(qa.quantity(tend, 'd')['value'], format='mjd')
    midtime = Time((starttim1.mjd + endtim1.mjd) / 2., format='mjd')

    tstr = midtime.to_datetime().strftime('{}_%Y%m%dT%H%M%S.%f'.format(observatory))

    if spw:
        spstr = 'spw{}'.format(spw.replace('~', '-'))
        filename = '.'.join([tstr, spstr])
    else:
        filename = tstr

    return filename


def msclearhistory(msfile):
    '''Clears history in the a measurement sets file

    :param msfile: string
            The name of a measurement sets file

    :return:
    '''

    tb.open(msfile + '/HISTORY', nomodify=False)
    nrows = tb.nrows()
    if nrows > 0:
        tb.removerows(range(nrows))
    tb.close()


def clearflagrow(msfile, mode='clear'):
    '''

    :param msfile:
    :param mode: FLAG_ROW operation
    default: 'clear': (default) clear the FLAG_ROW
             'list': to list existing FLAG_ROW
    :return:
    '''

    if mode == 'list':
        tb.open(msfile, nomodify=True)
        a = tb.getcol('FLAG_ROW')
        nfrows = np.sum(a)
        nrows = float(len(a))
        print('{:0d} out of {:.0f} ({:.0f}%) rows are flagged in {}'.format(nfrows, nrows, nfrows / nrows * 100,
                                                                            os.path.basename(msfile)))
    elif mode == 'clear':
        tb.open(msfile, nomodify=False)
        a = tb.getcol('FLAG_ROW')
        a[:] = False
        tb.putcol('FLAG_ROW', a)
        print('reset successfully')
    tb.close()


def splitX(vis, datacolumn2='MODEL_DATA', **kwargs):
    import os
    '''

    :param msfile:
    :param datacolumn:
    :param datacolumn2:
    :return:
    '''
    kwargs2 = kwargs.copy()
    datacolumn2 = datacolumn2.upper()

    outmsfile = kwargs['outputvis']
    if outmsfile.endswith('/'):
        outmsfile = outmsfile[:-1]
    if os.path.exists(outmsfile):
        os.system('rm -rf {}'.format(outmsfile))
    if os.path.exists('{}.flagversions'.format(outmsfile)):
        os.system('rm -rf {}.flagversions'.format(outmsfile))
    split(vis, **kwargs)

    for k in ['datacolumn', 'outputvis']:
        if k in kwargs2:
            kwargs2.pop(k)

    kwargs2['outputvis'] = 'tmpms.ms'
    kwargs2['datacolumn'] = datacolumn2.replace('_DATA', '')
    if os.path.exists('tmpms.ms'):
        os.system('rm -rf tmpms.ms')
    split(vis, **kwargs2)

    tb.open('tmpms.ms')
    nrows = tb.nrows()
    data = []
    for row in tqdm(range(nrows), desc='getting {} column'.format(datacolumn2), ascii=True):
        data.append(tb.getcell('DATA', row))
    tb.close()

    clearcal(outmsfile, addmodel=True)
    tb.open(outmsfile, nomodify=False)
    for row in tqdm(range(nrows), desc='writing {} column'.format(datacolumn2), ascii=True):
        tb.putcell(datacolumn2, row, data[row])
    tb.close()

    os.system('rm -rf tmpms.ms')
    return outmsfile


def flagcaltboutliers(caltable, limit=[]):
    import numpy as np
    import numpy.ma as ma
    # def removeOutliers(x, outlierConstant):
    #     a = np.array(x)
    #     idx, = np.where(np.diff(np.sort(datamag[0, 0, :]))>)
    #     upper_quartile = np.percentile(a, 80)
    #     lower_quartile = np.percentile(a, 20)
    #     IQR = (upper_quartile - lower_quartile) * outlierConstant
    #     quartileSet = (lower_quartile - IQR, upper_quartile + IQR)
    #     return ma.masked_outside(x, quartileSet[1], quartileSet[0])

    if not os.path.exists(caltable): return 0
    if isinstance(limit, list):
        if len(limit) == 2:
            tb.open(caltable, nomodify=False)
            # subt = tb.query("ANTENNA1==1 && SPECTRAL_WINDOW_ID=10")
            # data = subt.getcol('CPARAM')
            # flag = subt.getcol('FLAG')
            # spw = subt.getcol('SPECTRAL_WINDOW_ID')
            # datamag = np.abs(data)
            # mdatamag = ma.masked_outside(datamag, limit[0], limit[1])
            # mask = np.logical_or(mdatamag.mask, flag)
            # dataidx1 = datamag<limit[0]
            # dataidx2 = datamag>limit[1]
            # mdatamag = ma.masked_array(mdatamag, mask)
            # mdatamag[0, 0, :] = removeOutliers(mdatamag[0, 0, :], 5)
            # mdatamag[1, 0, :] = removeOutliers(mdatamag[1, 0, :], 5)
            data = tb.getcol('CPARAM')
            flag = tb.getcol('FLAG')
            datamag = np.abs(data)
            dataidx1 = datamag < limit[0]
            dataidx2 = datamag > limit[1]
            flag[dataidx1] = True
            flag[dataidx2] = True
            tb.putcol('FLAG', flag)
            return 1
        else:
            print('limit must have two elements. Aborted!')
            return 0
    else:
        print('limit must be a list. Aborted!')


def modeltransfer(msfile, spw='', reference='XX', transfer='YY'):
    from taskinit import mstool
    pol_dict = {'XX': 0, 'YY': 1, 'XY': 2, 'YX': 3}
    refidx = pol_dict[reference]
    trfidx = pol_dict[transfer]
    datams = mstool()
    datams.open(msfile, nomodify=False)

    if '~' in spw:
        sp0, sp1 = spw.split('~')
        for sp in range(int(sp0), int(sp1) + 1):
            staql = {'spw': str(sp)}
            datams.selectinit(reset=True)
            datams.msselect(staql)
            modeldata = datams.getdata(['model_data'])
            modeldata['model_data'][trfidx, ...] = modeldata['model_data'][refidx, ...]
            datams.putdata(modeldata)
        datams.close()
    else:
        datams.selectinit(reset=True)
        staql = {'spw': spw}
        datams.msselect(staql)
        modeldata = datams.getdata(['model_data'])
        modeldata['model_data'][trfidx, ...] = modeldata['model_data'][refidx, ...]
        datams.putdata(modeldata)
        datams.close()


def concat_slftb(tb_in=[], tb_out=None):
    if not tb_in:
        print('tb_in not provided. Abort...')
    if os.path.exists(tb_out):
        os.system('rm -r {}'.format(tb_out))
    os.system('cp -r {} {}'.format(tb_in[0], tb_out))
    tbdata = {}
    tb.open(tb_out)
    cols = tb.colnames()
    tb.close()
    cols.remove('WEIGHT')
    for col in cols:
        tbdata[col] = []
    for tbidx, ctb in enumerate(tb_in):
        tb.open(ctb, nomodify=True)
        tim0 = tb.getcol(cols[0])
        if len(tim0) == 0:
            continue
        else:
            for col in cols:
                if tbidx == 1 and col in ['CPARAM', 'PARAMERR', 'FLAG', 'SNR']:
                    tbdata[col].append(tb.getcol(col)[::-1, ...])
                else:
                    tbdata[col].append(tb.getcol(col))
        tb.close()

    if len(tbdata[cols[0]]) == 0:
        print('tables have no data. Return')
        return -1
    else:
        for col in cols:
            if col in ['CPARAM', 'PARAMERR', 'FLAG', 'SNR']:
                tbdata[col] = np.concatenate(tbdata[col], axis=2)
            else:
                tbdata[col] = np.concatenate(tbdata[col])
        tb.open(tb_out, nomodify=False)
        nrows = tb.nrows()
        nrows_new = len(tbdata[cols[0]])
        tb.addrows(nrows_new - nrows)
        for col in cols:
            tb.putcol(col, tbdata[col])
        tb.close()
        return tb_out


def gaincalXY(vis=None, caltable=None, pols='XXYY', msfileXY=None, gaintableXY=None, **kwargs):
    from gaincal_cli import gaincal_cli as gaincal
    if pols == 'XXYY':
        pols = 'XX,YY'
    pols_ = pols.split(',')
    rm_msfileXY = False
    if msfileXY is None:
        rm_msfileXY = True
        msfileXY = {}
        for pol in pols_:
            msfileXY[pol] = '.'.join([vis, pol])
            if os.path.exists(msfileXY[pol]):
                os.system('rm -rf {}'.format(msfileXY[pol]))
            splitX(vis=vis, outputvis=msfileXY[pol], correlation=pol, datacolumn='data', datacolumn2='MODEL_DATA')
    if gaintableXY is not None:
        if 'gaintable' in kwargs.keys():
            kwargs.pop('gaintable')
    caltbXY = []
    for pol in pols_:
        caltb_ = '.'.join([caltable, pol])
        if gaintableXY is not None:
            kwargs['gaintable'] = gaintableXY[pol]
        gaincal(vis=msfileXY[pol], caltable=caltb_, **kwargs)
        caltbXY.append(caltb_)
    concat_slftb(caltbXY, caltable)
    if rm_msfileXY:
        for k, v in msfileXY.iteritems():
            os.system('rm -rf {}'.format(v))
    return


def getmodel(vis, spw=None):
    tb.open(vis, nomodify=True)
    subt = tb.query("DATA_DESC_ID==" + str(spw))
    model_d = subt.getcol('MODEL_DATA')
    subt.done()
    tb.done()
    return model_d


def putmodel(vis, spw=None, model=None):
    tb.open(vis, nomodify=False)
    subt = tb.query("DATA_DESC_ID==" + str(spw))
    model_d = subt.putcol('MODEL_DATA', model)
    subt.done()
    tb.done()
    return model_d



import pyqtgraph as pg
import numpy as np
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import os
import pickle
from regions import Regions, CRTFRegionParserError
from PyQt5.QtGui import QDesktopServices

class roi_dialog(object):
    def __init__(self, img_size, cfreqs):
        self.world = False
        self.img_size = img_size
        self.cfreqs = cfreqs
        self.display_list = []
        # self.input_example = '[[{0}, {1}],[{2}, {3}]],[1.0, 18.0]'.format(int(img_size[0] * 0.3),
        #                                                                  int(img_size[1] * 0.3),
        #                                                                  int(img_size[0] * 0.7),
        #                                                                  int(img_size[1] * 0.7))
        PixSzeLt = [int(self.img_size[0] * 0.1), int(self.img_size[1] * 0.1), int(self.img_size[0] * 0.3),
                    int(self.img_size[1] * 0.3), int(self.img_size[0] * 0.5), int(self.img_size[1] * 0.5)]
        self.input_examples = []
        self.input_examples.append(
            'box[[{0}pix, {1}pix], [{2}pix, {3}pix]], range=[1.0GHz, 18.0GHz]'.format(PixSzeLt[2], PixSzeLt[3],
                                                                                      PixSzeLt[0], PixSzeLt[1]))
        self.input_examples.append(
            'centerbox[[{0}pix, {1}pix], [{2}pix, {3}pix]], range=[1.0GHz, 18.0GHz]'.format(PixSzeLt[4], PixSzeLt[5],
                                                                                            PixSzeLt[0], PixSzeLt[1]))
        self.input_examples.append(
            'rotbox[[{0}pix, {1}pix], [{2}pix, {3}pix], 15deg], range=[1.0GHz, 18.0GHz]'.format(PixSzeLt[2],
                                                                                                PixSzeLt[3],
                                                                                                PixSzeLt[0],
                                                                                                PixSzeLt[1]))
        self.input_examples.append(
            'ellipse[[{0}pix, {1}pix], [{2}pix, {3}pix], 15deg], range=[1.0GHz, 18.0GHz]'.format(PixSzeLt[4],
                                                                                                 PixSzeLt[5],
                                                                                                 PixSzeLt[0],
                                                                                                 PixSzeLt[0] * 0.75))
        self.input_examples.append(
            'circle[[{0}pix, {1}pix], {2}pix], range=[1.0GHz, 18.0GHz]'.format(PixSzeLt[4], PixSzeLt[5], PixSzeLt[0]))
        self.input_example = self.input_examples[0]

    def setupUi(self, Dialog):
        Dialog.setObjectName("Manually Defined ROI(s)")
        Dialog.resize(600, 302)
        self.ok_cancel = QDialogButtonBox(Dialog)
        self.ok_cancel.setGeometry(QRect(500, 40, 81, 71))
        self.ok_cancel.setOrientation(Qt.Vertical)
        self.ok_cancel.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.ok_cancel.setObjectName("ok_cancel")
        self.rois_list_view = QListView(Dialog)
        self.rois_list_view.setGeometry(QRect(10, 10, 481, 200))
        self.rois_list_view.setObjectName("rois_list_view")
        self.slm = QStringListModel()
        self.slm.setStringList(self.display_list)
        self.selected_index = [cur_item.row() for cur_item in self.rois_list_view.selectedIndexes()]

        # listView.clicked.connect(self.checkItem)
        self.rois_list_view.setModel(self.slm)
        self.lineEdit = QLineEdit(Dialog)
        self.lineEdit.setGeometry(QRect(10, 220, 561, 21))
        self.lineEdit.setObjectName("lineEdit")
        self.lineEdit.setText(self.input_example)
        # self.label = QLabel(Dialog)
        # self.label.setGeometry(QRect(10, 160, 381, 21))
        # self.label.setObjectName("label")
        self.pushButton_url = QPushButton(Dialog)
        self.pushButton_url.setGeometry(QRect(500, 180, 81, 32))
        self.pushButton_url.setObjectName("pushButton_url")
        self.pushButton_url.clicked.connect(lambda: QDesktopServices.openUrl(
            QUrl("https://casaguides.nrao.edu/index.php/CASA_Region_Format#Global_definitions")))
        self.pushButton_delete = QPushButton(Dialog)
        self.pushButton_delete.setGeometry(QRect(500, 140, 81, 32))
        self.pushButton_delete.setObjectName("pushButton_delete")
        self.pushButton_delete.clicked.connect(self.delete_from_list)
        # self.checkBox = QCheckBox(Dialog)
        # self.checkBox.setGeometry(QRect(10, 250, 87, 20))
        # self.checkBox.setObjectName("checkBox")
        self.pushButton_add_to_list = QPushButton(Dialog)
        self.pushButton_add_to_list.setGeometry(QRect(470, 250, 113, 32))
        self.pushButton_add_to_list.setObjectName("pushButton")
        self.pushButton_add_to_list.clicked.connect(self.add_to_list)
        self.pushButton_img_flx_preset = QPushButton(Dialog)
        self.pushButton_img_flx_preset.setGeometry(QRect(10, 250, 120, 32))
        self.pushButton_img_flx_preset.setObjectName("img_flx_roi")
        self.pushButton_img_flx_preset.clicked.connect(self.rois_for_cal_flux)
        self.pushButton_load_file = QPushButton(Dialog)
        self.pushButton_load_file.setGeometry(QRect(200, 250, 140, 32))
        self.pushButton_load_file.setObjectName("load_roi_file")
        self.pushButton_load_file.clicked.connect(self.roi_file_select)
        self.shape_comboBox = QComboBox(Dialog)
        self.shape_comboBox.setGeometry(QRect(350, 250, 104, 26))
        self.shape_comboBox.setObjectName("comboBox")
        self.shape_comboBox.addItems([''] * 7)
        # self.label_2 = QLabel(Dialog)
        # self.label_2.setGeometry(QRect(10, 190, 221, 16))
        # self.label_2.setObjectName("label_2")

        self.retranslateUi(Dialog)
        self.ok_cancel.accepted.connect(Dialog.accept)
        self.ok_cancel.rejected.connect(Dialog.reject)
        QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Manually Define ROI(s)"))
        # self.label.setText(
        #    _translate("Dialog", "FOV in pixel/arcsec, freq_boundry in GHz"))
        # self.checkBox.setText(_translate("Dialog", "World"))
        self.pushButton_url.setText(_translate("Dialog", "Doc"))
        self.pushButton_img_flx_preset.setText(_translate("Dialog", "EOVSA Presets"))
        self.pushButton_load_file.setText(_translate("Dialog", "Load ROI File"))
        self.pushButton_add_to_list.setText(_translate("Dialog", "Add to List"))
        self.pushButton_delete.setText(_translate("Dialog", "Delete"))
        self.shape_comboBox.setItemText(0, _translate("Dialog", "box"))
        self.shape_comboBox.setItemText(1, _translate("Dialog", "centerbox"))
        self.shape_comboBox.setItemText(2, _translate("Dialog", "rotbox"))
        self.shape_comboBox.setItemText(3, _translate("Dialog", "ellipse"))
        self.shape_comboBox.setItemText(4, _translate("Dialog", "circle"))
        self.shape_comboBox.setItemText(5, _translate("Dialog", "poly"))
        self.shape_comboBox.setItemText(6, _translate("Dialog", "annulus"))
        self.shape_comboBox.currentIndexChanged.connect(self.change_example)
        # self.ok_cancel.accepted.connect(Dialog.accept)
        # self.ok_cancel.rejected.connect(Dialog.reject)
        # self.label_2.setText(_translate("Dialog", "e.g. [[x1,y1],[x2,y2]],[freq_lo,freq_hi]"))

    def add_to_list(self):
        self.input_string = self.lineEdit.text()
        self.cur_info_string = self.input_string
        try:
            tmp_crtf_region = Regions.parse(self.cur_info_string, format='crtf')[0]
            self.display_list.append(self.cur_info_string)
            self.slm.setStringList(self.display_list)
            self.rois_list_view.setModel(self.slm)
        except:
            msg_box = QMessageBox(QMessageBox.Warning, 'Invalid Input!', 'The input is not a vliad CRTF string!')
            msg_box.exec_()

    def delete_from_list(self):
        self.selected_sting = [cur_item.row() for cur_item in self.rois_list_view.selectedIndexes()]
        del self.display_list[self.selected_sting[0]]
        self.slm.setStringList(self.display_list)
        self.rois_list_view.setModel(self.slm)

    def change_example(self):
        self.input_example = self.input_examples[self.shape_comboBox.currentIndex()]
        self.lineEdit.setText(self.input_example)

    # @staticmethod
    def getResult(self, Dialog):
        # dialog = roi_dialog(parent)
        # cur_result = dialog.exec_()
        return self.display_list

    def rois_for_cal_flux(self):
        selections = [str(x) for x in np.arange(2, 6).tolist()]
        seleted_item, okPressed = QInputDialog.getItem(None, "Number of ROIs", "Number of ROIs", selections, 0, False)
        num_of_rois = int(seleted_item)
        print(self.cfreqs)
        if okPressed:
            print(num_of_rois, ' ROIs are created')
            freq_boundry = np.linspace(0, len(self.cfreqs) - 1, num_of_rois + 1, dtype=int)
            size_list = [int(max(100.0 * self.cfreqs[0] / self.cfreqs[freq_ind], 5.)) for freq_ind in
                         freq_boundry[:-1]]
            crtf_str_list = []
            for nri in range(num_of_rois):
                crtf_str_list.append(
                    'centerbox[[{}pix, {}pix], [{}pix, {}pix]], range=[{}GHz, {}GHz]'.format(int(self.img_size[0] / 2),
                                                                                             int(self.img_size[1] / 2),
                                                                                             size_list[nri],
                                                                                             size_list[nri],
                                                                                             self.cfreqs[freq_boundry[nri]],
                                                                                             self.cfreqs[freq_boundry[nri + 1]]))
            for cur_str in crtf_str_list:
                self.display_list.append(cur_str)
            self.slm.setStringList(self.display_list)
            self.rois_list_view.setModel(self.slm)
        else:
            return
    def roi_file_select(self, Dialog):
        cur_file_name, _file_filter = QFileDialog.getOpenFileName(None, 'Select Region save',
                                                                     './', 'Region save (*.crtf *.reg *.ds9 *.fits)')
        #self.fname = 'EOVSA_20210507T190205.000000.outim.image.allbd.fits'
        cur_format = cur_file_name.split('.')[-1]
        if cur_format == 'reg':
            cur_format = 'ds9'
        try:
            cur_region = Regions.read(cur_file_name, format=cur_file_name.split('.')[-1])
        except:
            msg_box = QMessageBox(QMessageBox.Warning, 'Invalid Input!', 'The input can not be converted!')
            msg_box.exec_()
        self.display_list.append(cur_region.serialize(format='crtf'))
        self.slm.setStringList(self.display_list)
        self.rois_list_view.setModel(self.slm)
        return



def crtf_to_pgroi(crtf_str, eo_wcs, pen_arg):
    try:
        crtf_region = Regions.parse(crtf_str, format='crtf')[0]
    except:
        print(
            'check your input string format: https://casaguides.nrao.edu/index.php/CASA_Region_Format#Global_definitions')
    if 'Sky' in str(type(crtf_region)):
        crtf_region = crtf_region.to_pixel(eo_wcs)

    def get_corner_rotate(center_xy, width, height, rot_angle):
        distance = np.sqrt(width ** 2 + height ** 2)
        cur_theta = np.pi / 2. - np.arctan(width / height) - rot_angle
        cor_x = center_xy[0] - distance * np.sin(cur_theta)
        cor_y = center_xy[1] - distance * np.cos(cur_theta)
        return [cor_x, cor_y]

    print(str(type(crtf_region)))
    if 'CirclePixel' in str(type(crtf_region)):
        pg_roi = pg.CircleROI(pos=np.asarray(crtf_region.center.xy) - np.ones(2) * crtf_region.radius,
                              radius=crtf_region.radius, pen=pen_arg)
    elif 'RectanglePixel' in str(type(crtf_region)):
        pg_roi = pg.RectROI(
            pos=get_corner_rotate(center_xy=crtf_region.center.xy, width=crtf_region.width, height=crtf_region.height,
                                  rot_angle=crtf_region.angle.to_value(unit='radian')),
            size=[crtf_region.width, crtf_region.height], angle=crtf_region.angle.to_value(unit='degree'), pen=pen_arg)
    elif 'EllipsePixel' in str(type(crtf_region)):
        pg_roi = pg.EllipseROI(pos=get_corner_rotate(center_xy=crtf_region.center.xy, width=crtf_region.width,
                                                     height=crtf_region.height,
                                                     rot_angle=crtf_region.angle.to_value(unit='radian')),
                               size=[crtf_region.width, crtf_region.height],
                               angle=crtf_region.angle.to_value(unit='degree'), pen=pen_arg)
    # todo: add line ROI
    else:
        raise NameError('Sorry, only rectangle, circle, and ellipse are supported for this moment.')
    freq_range = [1.0, 18.0]
    if 'range' in crtf_region.meta:
        if str(crtf_region.meta['range'][0].unit) == 'GHz':
            freq_range = [crtf_region.meta['range'][0].value, crtf_region.meta['range'][1].value]
            print('Set freq range to {}'.format(freq_range))
    return (pg_roi, freq_range)

def rois_for_cal_flux(self1):
    selections = [str(x) for x in np.arange(2, 6).tolist()]
    seleted_item, okPressed = QInputDialog.getItem(None, "Number of ROIs", "Number of ROIs", selections, 0, False)
    num_of_rois = int(seleted_item)
    if okPressed:
        print(num_of_rois, ' ROIs are created')
        add_new_group(self1)
        freq_boundry = np.linspace(0, len(self1.cfreqs) - 1, num_of_rois + 1, dtype=int)
        size_list = [int(max(100.0 * self1.cfreqs[0] / self1.cfreqs[freq_ind], 5.)) for freq_ind in freq_boundry[:-1]]
        crtf_str_list = []
        for nri in range(num_of_rois):
            crtf_str_list.append(
                'centerbox[[{}pix, {}pix], [{}pix, {}pix]], range=[{}GHz, {}GHz]'.format(int(self1.meta['nx'] / 2),
                                                                                         int(self1.meta['ny'] / 2),
                                                                                         size_list[nri], size_list[nri],
                                                                                         freq_boundry[nri],
                                                                                         freq_boundry[nri + 1]))
        add_md_rois(self1,crtf_str_list)
    else:
        print('No ROI is defined.')


def save_roi_group(self1):
    data_saved = False
    selections = []
    for group_index, cur_group in enumerate(self1.rois):
        selections.append('Group {}, {} ROI(s)'.format(group_index, len(cur_group)))
    seleted_item, okPressed = QInputDialog.getItem(None, "Which Group?", "List of Group", selections, 0, False)
    if okPressed:
        print(seleted_item, ' is selected')
        seleted_group_index = int(seleted_item.split(' ')[1][:-1])
        fileName, ok2 = QFileDialog.getSaveFileName(None,
                                                    "Save the selected group ()",
                                                    os.getcwd(),
                                                    "All Files (*);;pickle save (*.p)")
        if ok2:
            with open(fileName, 'wb') as handle:
                pickle.dump([cur_roi.saveState() for cur_roi in self1.rois[seleted_group_index]], handle,
                            protocol=pickle.HIGHEST_PROTOCOL)
                print('The selected group has been saved to ', fileName)
                data_saved = True
    if not data_saved:
        print('No ROI is saved.')


def add_md_rois(self1, inp_str_list):
    add_new_group(self1)
    for si, cur_str in enumerate(inp_str_list):
        add_customized_roi(self1=self1, crtf_str=cur_str)


def add_customized_roi(self1, crtf_str):
    """Add a ROI region to the selection"""
    pg_roi_obj, freq_range = crtf_to_pgroi(crtf_str=crtf_str, eo_wcs=self1.eo_wcs,
                                           pen_arg=(len(self1.rois[self1.roi_group_idx]), 9))
    self1.new_roi = pg_roi_obj
    self1.pg_img_canvas.addItem(self1.new_roi)
    self1.new_roi.freq_mask = np.ones_like(self1.cfreqs) * False
    self1.new_roi.sigRegionChanged.connect(self1.calc_roi_spec)
    # choose which group to add
    self1.add_to_roigroup_selection()
    self1.rois[self1.roi_group_idx].append(self1.new_roi)
    self1.nroi_current_group = len(self1.rois[self1.roi_group_idx])
    self1.roi_selection_widget.clear()
    self1.roi_selection_widget.addItems([str(i) for i in range(self1.nroi_current_group)])
    self1.current_roi_idx = self1.nroi_current_group - 1
    self1.has_rois = True
    self1.calc_roi_spec()
    self1.roi_freq_lowbound_selector.setValue(freq_range[0])
    self1.roi_freq_hibound_selector.setValue(freq_range[1])


def add_new_group(self):
    if self.has_rois:
        self.rois.append([])
        self.roi_group_idx += 1
        if len(self.rois) > self.add_to_roigroup_widget.count():
            self.add_to_roigroup_widget.addItem(str(self.roi_group_idx))
            self.roigroup_selection_widget.addItem(str(self.roi_group_idx))
    else:
        self.roi_group_idx = 0
    self.add_to_roigroup_widget.setCurrentRow(self.roi_group_idx)
    self.add_to_roigroup_button.setText(self.add_to_roigroup_widget.currentItem().text())
    self.roigroup_selection_widget.setCurrentRow(self.roi_group_idx)
    self.roigroup_selection_button.setText(self.roigroup_selection_widget.currentItem().text())


##################### generated by xml-casa (v2) from importeovsa.xml ###############
##################### 27a519e51ee300bf7df502106dade919 ##############################
from __future__ import absolute_import
import numpy
from casatools.typecheck import CasaValidator as _val_ctor
_pc = _val_ctor( )
from casatools.coercetype import coerce as _coerce
from .private.task_importeovsa import importeovsa as _importeovsa_t
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log

class _importeovsa:
    """
    importeovsa ---- Parallelized import EOVSA idb file(s) to a measurement set or multiple measurement set.

    
    Parallelized imports an arbitrary number of EOVSA idb-format data sets into
    a casa measurement set. If more than one band is present, they
    will be put in the same measurement set but in a separate spectral
    window.
    

    --------- parameter descriptions ---------------------------------------------

    idbfiles          Name of input EOVSA idb file(s) or observation time range.
    ncpu              Number of cpu cores to use
    timebin           Bin width for time averaging
    width             Width of output channel relative to MS channel (# to average)
    visprefix         Prefix of vis names (may include the path).
    udb_corr          if applying correction to input UDB files before import to MS.
    use_exist_udbcorr If use the existed udb_corr results.
    nocreatms         If setting nocreatms True, will simulate a model measurement set for the first idb file and copy the model for the rest of idl files in list. If False, will simulate a new measurement set for every idbfile in list.
    doconcat          If concatenate multi casa measurement sets to one file.
    modelms           Name of input model measurement set file. If modelms is assigned, no simulation will start.
    doscaling         If creating a new MS file with the amplitude of visibility data rescaled.
    keep_nsclms       Keep the no scaling measurement sets

    --------- examples -----------------------------------------------------------

    
    
    Parallelized imports an arbitrary number of EOVSA idb-format data sets into
    a casa measurement set. If more than one band is present, they
    will be put in the same measurement set but in a separate spectral
    window.
    
    Detailed Keyword arguments:
    
    idbfiles -- Name of input EOVSA idb file(s)
    default: none. Must be supplied
    example: idbfiles = 'IDB20160524000518'
    example: idbfiles=['IDB20160524000518','IDB20160524000528']
    
    ncpu -- Number of cpu cores to use
    default: 8
    
    visprefix -- Prefix of vis names (may include the path)
    default: none;
    example: visprefix='sun/']
    
    
    --- Data Selection ---
    
    nocreatms -- If copying a new MS file instead of create one from MS simulator.
    default: False
    
    modelms -- Name of the standard Measurement Set. IF modelms is not provided, use
    '/home/user/sjyu/20160531/ms/sun/SUN/SUN_20160531T142234-10m.1s.ms' as a standard MS.
    
    doconcat -- If outputing one single MS file
    
    
    --- Channel averaging parameter ---
    
    width -- Number of input channels to average to create an output
    channel. If a list is given, each bin will apply to one spw in
    the selection.
    default: 1 => no channel averaging.
    options: (int) or [int]
    
    example: chanbin=[2,3] => average 2 channels of 1st selected
    spectral window and 3 in the second one.
    
    
    --- Time averaging parameters ---
    
    timebin -- Bin width for time averaging. When timebin is greater than 0s,
    the task will average data in time. Flagged data will be included
    in the average calculation, unless the parameter keepflags is set to False.
    In this case only partially flagged rows will be used in the average.
    default: '0s'
    
    
    
    
    
    
    
    
    
    


    """

    _info_group_ = """Import/export"""
    _info_desc_ = """Parallelized import EOVSA idb file(s) to a measurement set or multiple measurement set."""

    def __call__( self, idbfiles='', ncpu=int(1), timebin='0s', width=int(1), visprefix='', udb_corr=True, use_exist_udbcorr=False, nocreatms=False, doconcat=False, modelms='', doscaling=False, keep_nsclms=False ):
        schema = {'idbfiles': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'ncpu': {'type': 'cInt'}, 'timebin': {'type': 'cStr', 'coerce': _coerce.to_str}, 'width': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}, {'type': 'cInt'}, {'type': 'cIntVec', 'coerce': [_coerce.to_list,_coerce.to_intvec]}]}, 'visprefix': {'type': 'cStr', 'coerce': _coerce.to_str}, 'udb_corr': {'type': 'cBool'}, 'use_exist_udbcorr': {'type': 'cBool'}, 'nocreatms': {'type': 'cBool'}, 'doconcat': {'type': 'cBool'}, 'modelms': {'type': 'cStr', 'coerce': _coerce.to_str}, 'doscaling': {'type': 'cBool'}, 'keep_nsclms': {'type': 'cBool'}}
        doc = {'idbfiles': idbfiles, 'ncpu': ncpu, 'timebin': timebin, 'width': width, 'visprefix': visprefix, 'udb_corr': udb_corr, 'use_exist_udbcorr': use_exist_udbcorr, 'nocreatms': nocreatms, 'doconcat': doconcat, 'modelms': modelms, 'doscaling': doscaling, 'keep_nsclms': keep_nsclms}
        assert _pc.validate(doc,schema), str(_pc.errors)
        _logging_state_ = _start_log( 'importeovsa', [ 'idbfiles=' + repr(_pc.document['idbfiles']), 'ncpu=' + repr(_pc.document['ncpu']), 'timebin=' + repr(_pc.document['timebin']), 'width=' + repr(_pc.document['width']), 'visprefix=' + repr(_pc.document['visprefix']), 'udb_corr=' + repr(_pc.document['udb_corr']), 'use_exist_udbcorr=' + repr(_pc.document['use_exist_udbcorr']), 'nocreatms=' + repr(_pc.document['nocreatms']), 'doconcat=' + repr(_pc.document['doconcat']), 'modelms=' + repr(_pc.document['modelms']), 'doscaling=' + repr(_pc.document['doscaling']), 'keep_nsclms=' + repr(_pc.document['keep_nsclms']) ] )
        return _end_log( _logging_state_, 'importeovsa', _importeovsa_t( _pc.document['idbfiles'], _pc.document['ncpu'], _pc.document['timebin'], _pc.document['width'], _pc.document['visprefix'], _pc.document['udb_corr'], _pc.document['use_exist_udbcorr'], _pc.document['nocreatms'], _pc.document['doconcat'], _pc.document['modelms'], _pc.document['doscaling'], _pc.document['keep_nsclms'] ) )

importeovsa = _importeovsa( )




import sys
py3 = sys.version_info.major >= 3

if py3:
    from casatools import table as tbtool
    from casatools import ms as mstool
    from casatools import quanta as qatool
    tb = tbtool()
    ms = mstool()
    qa = qatool()
else:
    from taskinit import tb, qa, ms


def getAntennaPosition(vis):
    tb.open(vis + '/ANTENNA')
    position = tb.getcol('POSITION')
    diameter = tb.getcol('DISH_DIAMETER')
    antenna = tb.getcol('NAME')
    tb.close()
    return position, diameter, antenna


def getObservatoryName(ms):
    """
    Returns the observatory name in the specified ms, using the tb tool.
    -- Todd Hunter
    """
    antTable = ms + '/OBSERVATION'
    try:
        tb.open(antTable)
        myName = tb.getcell('TELESCOPE_NAME')
        tb.close()
    except:
        print("Could not open OBSERVATION table to get the telescope name: {}".format(antTable))
        myName = ''
    return (myName)


def buildConfigurationFile(vis='', cfgfile=None):
    observatory = getObservatoryName(vis)
    position, diameter, antenna = getAntennaPosition(vis)
    if cfgfile == '':
        cfgfile = vis + '.cfg'
    cfg = open(cfgfile, 'w')
    coordsys = '# coordsys=XYZ\n'
    cfg.write('# observatory={}\n'.format(observatory))
    cfg.write(coordsys)
    for s in range(len(antenna)):
        x, y, z = position[:, s]
        line = '{:.6f} {:.6f} {:.6f} {:5.1f}  {}\n'.format(x, y, z, diameter[s], antenna[s])
        cfg.write(line)
    cfg.close()
    return cfgfile



import os
import numpy as np
from datetime import datetime
from math import *
# import jdutil
import bisect
import pdb
from taskinit import *

try:
    from astropy.io import fits as pyfits
except:
    try:
        import pyfits
    except ImportError:
        raise ImportError('Neither astropy nor pyfits exists in this CASA installation')


# from astropy.constants import R_sun, au

def read_horizons(ephemfile=None):
    # inputs:
    #   ephemfile:
    #       OBSERVER output from JPL Horizons for topocentric coordinates with for example
    #       target=Sun, observer=VLA=-5
    #       extra precision, quantities 1,17,20, REFRACTION
    #       routine goes through file to find $$SOE which is start of ephemeris and ends with $$EOE
    # outputs: a Python dictionary containing the following:
    #   timestr: date and time as a string
    #   time: modified Julian date
    #   ra: right ascention, in rad
    #   dec: declination, in rad
    #   rastr: ra in string
    #   decstr: dec in string
    #   p0: solar p angle, CCW with respect to the celestial north pole
    #   delta: distance from the disk center to the observer, in AU
    #   delta_dot: time derivative of delta, in the light of sight direction. Negative means it is moving toward the observer
    #
    if not ephemfile or ephemfile.isspace():
        raise ValueError, 'Please specify input ephem file!'
    if not os.path.isfile(ephemfile):
        raise ValueError, 'The specified input ephem file does not exist!'
    # initialize the return dictionary
    ephem0 = dict.fromkeys(['time', 'timestr', 'ra', 'dec', 'rastr', 'decstr', 'delta', 'delta_dot', 'p0'])
    f = open(ephemfile, 'rU')  # force the universal newline mode
    lines = f.readlines()
    f.close()
    nline = len(lines)
    for i in range(nline):
        line = lines[i]
        if line[0:5] == '$$SOE':  # start recording
            istart = i + 1
        if line[0:5] == '$$EOE':  # end recording
            iend = i
    newlines = lines[istart:iend]
    nrec = len(newlines)
    ephem_ = []
    for j in range(nrec):
        line = newlines[j]
        dt = datetime.strptime(line[1:18], '%Y-%b-%d %H:%M')
        newdtstr = dt.strftime('%Y-%m-%dT%H:%M')
        ephem0['timestr'] = newdtstr
        mjd = qa.quantity(newdtstr)
        # jd=jdutil.datetime_to_jd(dt)
        # mjd=jdutil.jd_to_mjd(jd)
        ephem0['time'] = mjd
        ephem0['rastr'] = line[23:36]
        ephem0['decstr'] = line[37:50]
        ephem0['ra'] = {'unit': 'rad', 'value': radians(
            (long(line[23:25]) + long(line[26:28]) / 60. + float(line[29:36]) / 3600.) * 15.)}  # in rad
        ephem0['dec'] = {'unit': 'rad', 'value': radians(
            long(line[38:40]) + long(line[41:43]) / 60. + float(line[44:50]) / 3600.)}  # in rad
        ephem0['p0'] = {'unit': 'deg', 'value': float(line[51:59])}
        ephem0['delta'] = {'unit': 'au', 'value': float(line[70:86])}
        ephem0['delta_dot'] = {'unit': 'km/s', 'value': float(line[88:98])}
        if line[37] == '-':
            ephem0['dec']['value'] = -ephem0['dec']['value']
        ephem_.append(ephem0.copy())
    # convert list of dictionary to a dictionary of arrays
    times = [ep['time']['value'] for ep in ephem_]
    ras = [ep['ra']['value'] for ep in ephem_]
    decs = [ep['dec']['value'] for ep in ephem_]
    p0s = [ep['p0']['value'] for ep in ephem_]
    deltas = [ep['delta']['value'] for ep in ephem_]
    ephem = {'times': times, 'ras': ras, 'decs': decs, 'p0s': p0s, 'deltas': deltas}
    return ephem


def read_msinfo(msfile=None, msinfofile=None):
    # read MS information #
    msinfo = dict.fromkeys(['msfile', 'scans', 'fieldids', 'btimes', 'btimestr', 'inttimes', 'ras', 'decs'])
    ms.open(msfile)
    scans = ms.getscansummary()
    scanids = sorted(scans.keys(), key=lambda x: int(x))
    nscanid = len(scanids)
    btimes = []
    btimestr = []
    etimes = []
    fieldids = []
    inttimes = []
    dirs = []
    ras = []
    decs = []
    for i in range(nscanid):
        btimes.append(scans[scanids[i]]['0']['BeginTime'])
        etimes.append(scans[scanids[i]]['0']['EndTime'])
        fieldid = scans[scanids[i]]['0']['FieldId']
        fieldids.append(fieldid)
        dir = ms.getfielddirmeas('PHASE_DIR', fieldid)
        dirs.append(dir)
        ras.append(dir['m0'])
        decs.append(dir['m1'])
        inttimes.append(scans[scanids[i]]['0']['IntegrationTime'])
    ms.close()
    btimestr = [qa.time(qa.quantity(btimes[i], 'd'), form='fits', prec=10)[0] for i in range(nscanid)]
    msinfo['msfile'] = msfile
    msinfo['scans'] = scans
    msinfo['fieldids'] = fieldids
    msinfo['btimes'] = btimes
    msinfo['btimestr'] = btimestr
    msinfo['inttimes'] = inttimes
    msinfo['ras'] = ras
    msinfo['decs'] = decs
    if msinfofile:
        np.savez(msinfofile, msfile=msfile, scans=scans, fieldids=fieldids, btimes=btimes, btimestr=btimestr,
                 inttimes=inttimes, ras=ras, decs=decs)
    return msinfo


def ephem_to_helio(msinfo=None, ephem=None, reftime=None, polyfit=None):
    ## 1. Take a solar ms database, read the scan and field information, find out the pointings (in RA and DEC)
    ## 2. Compare with the ephemeris of the solar disk center (in RA and DEC)
    ## 3. Generate VLA pointings in heliocentric coordinates
    ## inputs:
    ##      msinfo: CASA MS information, output from read_msinfo
    ##      ephem: solar ephem, output from read_horizons
    ##      reftime: list of reference times (e.g., used for imaging)
    #                CASA standard time format, either a single time (e.g., '2012/03/03/12:00:00'
    ##               or a time range (e.g., '2012/03/03/12:00:00~2012/03/03/13:00:00'. If the latter,
    ##               take the midpoint of the timerange for reference. If no date specified, take
    ##               the date of the first scan
    #       polyfit: ONLY works for MS database with only one source with continously tracking; not recommanded unless scan length is too long and want to have very high accuracy
    ## return value:
    ##      helio: a list of VLA pointing information
    ##              reftimestr: reference time, in FITS format string
    ##              reftime: reference time, in mjd format
    ##              ra: actual RA of VLA pointing at the reference time (interpolated)
    ##              dec: actual DEC of VLA pointing at the reference time (interpolated)
    ##              # CASA knows only RA and DEC of the closest field (e.g. in clean) #
    ##              ra_fld: right ascention of the CASA reference pointing direction
    ##              dec_fld: declination of the CASA reference pointing direction
    ##              raoff: RA offset of the actual VLA pointing to solar center
    ##              decoff: DEC offset of the actual VLA pointing to solar center
    ##              refx: heliocentric X offset of the actual VLA pointing to solar center
    ##              refy: heliocentric Y offset of the actual VLA pointing to solar center
    ######## Example #########
    #    msfile='sun_C_20140910T221952-222952.10s.cal.ms'
    #    ephemfile='horizons_sun_20140910.radecp'
    #    ephem=vla_prep.read_horizons(ephemfile=ephemfile)
    #    msinfo=vla_prep.read_msinfo(msfile=msfile)
    #    polyfit=0
    #    reftime = '22:25:20~22:25:40'
    if not ephem:
        raise ValueError, 'Please provide information of the MS database!'
    if not msinfo:
        raise ValueError, 'Please provide information of the MS database!'
    if isinstance(msinfo, str):
        try:
            msinfo0 = np.load(msinfo)
        except:
            raise ValueError, 'The specified input msinfo file does not exist!'
    elif isinstance(msinfo, dict):
        msinfo0 = msinfo
    else:
        raise ValueError, 'msinfo should be either a numpy npz or a dictionary'
    print 'msinfo is derived from: ', msinfo0['msfile']
    scans = msinfo0['scans']
    fieldids = msinfo0['fieldids']
    btimes = msinfo0['btimes']
    inttimes = msinfo0['inttimes']
    ras = msinfo0['ras']
    decs = msinfo0['decs']
    ra_rads = [ra['value'] for ra in ras]
    dec_rads = [dec['value'] for dec in decs]
    # fit 2nd order polynomial fits to the RAs and DECs #
    if polyfit:
        cra = np.polyfit(btimes, ra_rads, 2)
        cdec = np.polyfit(btimes, dec_rads, 2)

    # find out pointing direction according to the input time or timerange #
    if not reftime:
        raise ValueError, 'Please specify a reference time for pointing/imaging!'
    if isinstance(reftime, str):
        reftime = [reftime]
    if (not isinstance(reftime, list)):
        print 'input "reftime" is not a valid list. Abort...'

    nreftime = len(reftime)
    helio = []
    for reftime0 in reftime:
        helio0 = dict.fromkeys(
            ['reftimestr', 'reftime', 'ra', 'dec', 'ra_fld', 'dec_fld', 'raoff', 'decoff', 'refx', 'refy', 'p0'])
        helio0['reftimestr'] = reftime0
        if '~' in reftime0:
            # if reftime0 is specified as a timerange
            [tbg0, tend0] = reftime0.split('~')
            tbg_d = qa.getvalue(qa.convert(qa.totime(tbg0), 'd'))[0]
            tend_d = qa.getvalue(qa.convert(qa.totime(tend0), 'd'))[0]
            tdur_s = (tend_d - tbg_d) * 3600. * 24.
            # if no date is specified, add up the date of the first scan
            if tend_d < 1.:
                if tend_d >= tbg_d:
                    tend_d += int(btimes[0])
                else:
                    tend_d += int(btimes[0]) + 1
            if tbg_d < 1.:
                tbg_d += int(btimes[0])
            tref_d = (tbg_d + tend_d) / 2.
        else:
            # if reftime0 is specified as a single value
            tref_d = qa.getvalue(qa.convert(qa.totime(reftime0), 'd'))
            # if no date is specified, add up the date of the first scan
            if tref_d < 1.:
                tref_d += int(btimes[0])
            tbg_d = tref_d
            # use the intergration time
            ind = bisect.bisect_left(btimes, tref_d)
            tdur_s = inttims[ind - 1]
        helio0['reftime'] = tref_d
        helio0['date-obs'] = qa.time(qa.quantity(tbg_d, 'd'), form='fits', prec=10)[0]
        helio0['exptime'] = tdur_s

        # find out RA and DEC coords according to the reference time
        # if polyfit, then use the 2nd order polynomial coeffs
        ind = bisect.bisect_left(btimes, tref_d)
        if ind > 1:
            dt = tref_d - btimes[ind - 1]
            if ind < len(btimes):
                scanlen = btimes[ind] - btimes[ind - 1]
                (ra_b, ra_e) = (ras[ind - 1]['value'], ras[ind]['value'])
                (dec_b, dec_e) = (decs[ind - 1]['value'], decs[ind]['value'])
            if ind >= len(btimes):
                scanlen = btimes[ind - 1] - btimes[ind - 2]
                (ra_b, ra_e) = (ras[ind - 2]['value'], ras[ind - 1]['value'])
                (dec_b, dec_e) = (decs[ind - 2]['value'], decs[ind - 1]['value'])
        if ind == 1:  # only one scan exists (e.g., imported from AIPS)
            ra_b = ras[ind - 1]['value']
            ra_e = ra_b
            dec_b = decs[ind - 1]['value']
            dec_e = dec_b
            scanlen = 10.  # radom value
            dt = 0.
        if ind < 1:
            raise ValueError, 'Reference time does not fall into the scan list!'
        if polyfit:
            ra = cra[0] * tref_d ** 2. + cra[1] * tref_d + cra[2]
            dec = cdec[0] * tref_d ** 2. + cdec[1] * tref_d + cdec[2]
        # if not, use linearly interpolated RA and DEC at the beginning of this scan and next scan
        else:
            ra = ra_b + (ra_e - ra_b) / scanlen * dt
            dec = dec_b + (dec_e - dec_b) / scanlen * dt
        if ra < 0:
            ra += 2. * np.pi
        if ra_b < 0:
            ra_b += 2. * np.pi

        # compare with ephemeris from JPL Horizons
        time0s = ephem['times']
        ra0s = ephem['ras']
        dec0s = ephem['decs']
        p0s = ephem['p0s']
        delta0s = ephem['deltas']
        ind = bisect.bisect_left(time0s, tref_d)
        dt0 = time0s[ind] - time0s[ind - 1]
        dt_ref = tref_d - time0s[ind - 1]
        dra0 = ra0s[ind] - ra0s[ind - 1]
        ddec0 = dec0s[ind] - dec0s[ind - 1]
        dp0 = p0s[ind] - p0s[ind - 1]
        ddelta0 = delta0s[ind] - delta0s[ind - 1]
        ra0 = ra0s[ind - 1] + dra0 / dt0 * dt_ref
        dec0 = dec0s[ind - 1] + ddec0 / dt0 * dt_ref
        p0 = p0s[ind - 1] + dp0 / dt0 * dt_ref
        delta0 = delta0s[ind - 1] + ddelta0 / dt0 * dt_ref
        if ra0 < 0:
            ra0 += 2. * np.pi

        # RA and DEC offset in arcseconds
        decoff = degrees((dec - dec0)) * 3600.
        raoff = degrees((ra - ra0) * cos(dec)) * 3600.
        # Convert into heliocentric offsets
        prad = -radians(p0)
        refx = (-raoff) * cos(prad) - decoff * sin(prad)
        refy = (-raoff) * sin(prad) + decoff * cos(prad)
        helio0['ra'] = ra  # ra of the actual pointing
        helio0['dec'] = dec  # dec of the actual pointing
        helio0['ra_fld'] = ra_b  # ra of the field, used as the reference in e.g., clean
        helio0['dec_fld'] = dec_b  # dec of the field, used as the refenrence in e.g., clean
        helio0['raoff'] = raoff
        helio0['decoff'] = decoff
        helio0['refx'] = refx
        helio0['refy'] = refy
        helio0['p0'] = p0
        # helio['r_sun']=np.degrees(R_sun.value/(au.value*delta0))*3600. #in arcsecs
        helio.append(helio0)
    return helio


def getbeam(imagefile=None, beamfile=None):
    if not imagefile:
        raise ValueError, 'Please specify input images'
    bmaj = []
    bmin = []
    bpa = []
    beamunit = []
    bpaunit = []
    chans = []
    nimg = len(imagefile)
    for n in range(nimg):
        img = imagefile[n]
        if not os.path.exists(img):
            raise ValueError, 'The input image does not exist!'
        ia.open(img)
        sum = ia.summary()
        bmaj_ = []
        bmin_ = []
        bpa_ = []
        if sum.has_key('perplanebeams'):  # beam vary with frequency
            nbeams = sum['perplanebeams']['nChannels']
            beams = sum['perplanebeams']['beams']
            chans_ = [key[1:] for key in beams.keys()]
            chans_.sort(key=float)
            for chan in chans_:
                bmaj0 = beams['*' + chan]['*0']['major']['value']
                bmaj_.append(bmaj0)
                bmin0 = beams['*' + chan]['*0']['minor']['value']
                bmin_.append(bmin0)
                bpa0 = beams['*' + chan]['*0']['positionangle']['value']
                bpa_.append(bpa0)
            beamunit_ = beams['*' + chans_[0]]['*0']['major']['unit']
            bpaunit_ = beams['*' + chans_[0]]['*0']['positionangle']['unit']
        if sum.has_key('restoringbeam'):  # only one beam
            bmaj_.append(sum['restoringbeam']['major']['value'])
            bmin_.append(sum['restoringbeam']['minor']['value'])
            bpa_.append(sum['restoringbeam']['positionangle']['value'])
            beamunit_ = sum['restoringbeam']['major']['unit']
            bpaunit_ = sum['restoringbeam']['positionangle']['unit']
            nbeams = 1
            chans_ = [0]

        bmaj.append(bmaj_)
        bmin.append(bmin_)
        bpa.append(bpa_)
        beamunit.append(beamunit_)
        bpaunit.append(bpaunit_)
        chans.append(chans_)
    if beamfile:  # write beams to ascii file
        print 'Writing beam info to ascii file...'
        f = open(beamfile, 'w')
        f.write('CHANNEL No., BMAJ (' + beamunit[0] + '), BMIN (' + beamunit[0] + '), BPA (' + bpaunit[0] + ')')
        f.write("\n")
        for n in range(nimg):
            f.write('----For image: ' + imagefile[n] + '----')
            f.write('\n')
            chans_ = chans[n]
            for i in range(len(chans_)):
                f.write(str(chans_[i]) + ', ' + str(bmaj[n][i]) + ', ' + str(bmin[n][i]) + ', ' + str(bpa[n][i]))
                f.write("\n")
        f.close()
    return bmaj, bmin, bpa, beamunit, bpaunit


def imreg(imagefile=None, fitsfile=None, beamfile=None, helio=None, offsetfile=None, toTb=None, scl100=None,
          verbose=False):
    if not imagefile:
        raise ValueError, 'Please specify input image'
    if not helio:
        raise ValueError, 'Please specify input coordinate info for image registration. Use ephem_to_helio to derive that info'
    if not fitsfile:
        fitsfile = [img + '.fits' for img in imagefile]
    # if len(imagefile) != len(helio):
    #    raise ValueError, 'Number of input images does not equal to number of helio coord headers!'
    if len(imagefile) != len(fitsfile):
        raise ValueError, 'Number of input images does not equal to number of output fits files!'
    # get restoring beam info
    (bmajs, bmins, bpas, beamunits, bpaunits) = getbeam(imagefile=imagefile, beamfile=beamfile)
    nimg = len(imagefile)
    if verbose:
        print str(nimg) + ' images to process...'
    for n in range(nimg):
        if verbose:
            print 'processing image #' + str(n)
        img = imagefile[n]
        fitsf = fitsfile[n]
        hel = helio[n]
        bmaj = bmajs[n]
        bmin = bmins[n]
        beamunit = beamunits[n]
        if not os.path.exists(img):
            raise ValueError, 'Please specify input image'
        if os.path.exists(fitsf):
            raise ValueError, 'Specified fits file already exists!'
        else:
            p0 = hel['p0']
            ia.open(img)
            imr = ia.rotate(pa=str(-p0) + 'deg')
            imr.tofits(fitsf, history=False)
            imr.close()
            sum = ia.summary()
            ia.close()
        # construct the standard fits header
        # RA and DEC of the reference pixel crpix1 and crpix2
        (imra, imdec) = (sum['refval'][0], sum['refval'][1])
        # find out the difference of the image center to the CASA reference center
        # RA and DEC difference in arcseconds
        ddec = degrees((imdec - hel['dec_fld'])) * 3600.
        dra = degrees((imra - hel['ra_fld']) * cos(hel['dec_fld'])) * 3600.
        # Convert into image heliocentric offsets
        prad = -radians(hel['p0'])
        dx = (-dra) * cos(prad) - ddec * sin(prad)
        dy = (-dra) * sin(prad) + ddec * cos(prad)
        if offsetfile:
            try:
                offset = np.load(offsetfile)
            except:
                raise ValueError, 'The specified offsetfile does not exist!'
            reftimes_d = offset['reftimes_d']
            xoffs = offset['xoffs']
            yoffs = offset['yoffs']
            timg_d = hel['reftime']
            ind = bisect.bisect_left(reftimes_d, timg_d)
            xoff = xoffs[ind - 1]
            yoff = yoffs[ind - 1]
        else:
            xoff = hel['refx']
            yoff = hel['refy']
        if verbose:
            print 'offset of image phase center to visibility phase center (arcsec): ', dx, dy
            print 'offset of visibility phase center to solar disk center (arcsec): ', xoff, yoff
        (crval1, crval2) = (xoff + dx, yoff + dy)
        # update the fits header to heliocentric coordinates

        hdu = pyfits.open(fitsf, mode='update')
        header = hdu[0].header
        (cdelt1, cdelt2) = (
            -header['cdelt1'] * 3600., header['cdelt2'] * 3600.)  # Original CDELT1, 2 are for RA and DEC in degrees
        header['cdelt1'] = cdelt1
        header['cdelt2'] = cdelt2
        header['cunit1'] = 'arcsec'
        header['cunit2'] = 'arcsec'
        header['crval1'] = crval1
        header['crval2'] = crval2
        header['ctype1'] = 'HPLN-TAN'
        header['ctype2'] = 'HPLT-TAN'
        header['date-obs'] = hel['date-obs']
        try:
            # this works for pyfits version of CASA 4.7.0 but not CASA 4.6.0
            header.update('exptime', hel['exptime'])
            header.update('p_angle', hel['p0'])
        except:
            # this works for astropy.io.fits
            header.append(('exptime', hel['exptime']))
            header.append(('p_angle', hel['p0']))

        # header.update('comment', 'Fits header updated to heliocentric coordinates by Bin Chen')

        # update intensity units, i.e. to brightness temperature?
        if toTb:
            data = hdu[0].data  # remember the data order is reversed due to the FITS convension
            dim = data.ndim
            sz = data.shape
            keys = header.keys()
            values = header.values()
            # which axis is frequency?
            faxis = keys[values.index('FREQ')][-1]
            faxis_ind = dim - int(faxis)
            if header['BUNIT'].lower() == 'jy/beam':
                header['BUNIT'] = 'K'
                for i in range(sz[faxis_ind]):
                    nu = header['CRVAL' + faxis] + header['CDELT' + faxis] * (i + 1 - header['CRPIX' + faxis])
                    if header['CUNIT' + faxis] == 'KHz':
                        nu *= 1e3
                    if header['CUNIT' + faxis] == 'MHz':
                        nu *= 1e6
                    if header['CUNIT' + faxis] == 'GHz':
                        nu *= 1e9
                    if len(bmaj) > 1:  # multiple (per-plane) beams
                        bmajtmp = bmaj[i]
                        bmintmp = bmin[i]
                    else:  # one single beam
                        bmajtmp = bmaj[0]
                        bmintmp = bmin[0]
                    if beamunit == 'arcsec':
                        bmaj0 = np.radians(bmajtmp / 3600.)
                        bmin0 = np.radians(bmajtmp / 3600.)
                    if beamunit == 'arcmin':
                        bmaj0 = np.radians(bmajtmp / 60.)
                        bmin0 = np.radians(bmintmp / 60.)
                    if beamunit == 'deg':
                        bmaj0 = np.radians(bmajtmp)
                        bmin0 = np.radians(bmintmp)
                    if beamunit == 'rad':
                        bmaj0 = bmajtmp
                        bmin0 = bmintmp
                    beam_area = bmaj0 * bmin0 * np.pi / (4. * log(2.))
                    k_b = qa.constants('k')['value']
                    c_l = qa.constants('c')['value']
                    factor = 2. * k_b * nu ** 2 / c_l ** 2  # SI unit
                    jy_to_si = 1e-26
                    # print nu/1e9, beam_area, factor
                    factor2 = 1.
                    if scl100:
                        factor2 = 100.
                    if faxis == '3':
                        data[:, i, :, :] *= jy_to_si / beam_area / factor * factor2
                    if faxis == '4':
                        data[i, :, :, :] *= jy_to_si / beam_area / factor * factor2

        hdu.flush()
        hdu.close()



"""
cookb_signalsmooth.py

from: http://scipy.org/Cookbook/SignalSmooth
"""

import numpy as np
import matplotlib.pyplot as plt

def smooth(x, window_len=10, window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    import numpy as np    
    t = np.linspace(-2,2,0.1)
    x = np.sin(t)+np.random.randn(len(t))*0.1
    y = smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    
    NOTE from B. Chen: slightly modified the reflected copies: window_len-1 points are added to both ends. 
    Previous one is not exactly the reflection, the indices are off by one pixel
    """

    if x.ndim != 1:
        raise(ValueError, "smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise(ValueError, "Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise(ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    #s=np.r_[2*x[0]-x[window_len:1:-1], x, 2*x[-1]-x[-1:-window_len:-1]]
    s=np.r_[2*x[0]-x[window_len-1:0:-1], x, 2*x[-1]-x[-2:-window_len-1:-1]]
    #print(len(s))
    
    if window == 'flat': #moving average
        w = np.ones(window_len,'d')
    else:
        w = getattr(np, window)(window_len)
    y = np.convolve(w/w.sum(), s, mode='same')
    return y[window_len-1:-window_len+1]


#*********** part2: 2d

from scipy import signal

def gauss_kern(size, sizey=None):
    """ Returns a normalized 2D gauss kernel array for convolutions """
    size = int(size)
    if not sizey:
        sizey = size
    else:
        sizey = int(sizey)
    x, y = np.mgrid[-size:size+1, -sizey:sizey+1]
    g = np.exp(-(x**2/float(size) + y**2/float(sizey)))
    return g / g.sum()

def blur_image(im, n, ny=None) :
    """ blurs the image by convolving with a gaussian kernel of typical
        size n. The optional keyword argument ny allows for a different
        size in the y direction.
    """
    g = gauss_kern(n, sizey=ny)
    improc = signal.convolve(im, g, mode='valid')
    return(improc)


def smooth_demo():
    t = np.linspace(-4,4,100)
    x = np.sin(t)
    xn = x + np.random.randn(len(t)) * 0.1
    y = smooth(x)
    ws = 31

    plt.subplot(211)
    plt.plot(np.ones(ws))

    windows=['flat', 'hanning', 'hamming', 'bartlett', 'blackman']

    plt.hold(True)
    for w in windows[1:]:
        #eval('plt.plot('+w+'(ws) )')
        plt.plot(getattr(np, w)(ws))

    plt.axis([0,30,0,1.1])

    plt.legend(windows)
    plt.title("The smoothing windows")
    plt.subplot(212)
    plt.plot(x)
    plt.plot(xn)
    for w in windows:
        plt.plot(smooth(xn,10,w))
    l = ['original signal', 'signal with noise']
    l.extend(windows)
    plt.legend(l)
    plt.title("Smoothing a noisy signal")
    #plt.show()


if __name__=='__main__':
    
    # part 1: 1d
    smooth_demo()
    
    # part 2: 2d
    X, Y = np.mgrid[-70:70, -70:70]
    Z = np.cos((X**2+Y**2)/200.)+ np.random.normal(size=X.shape)
    Z2 = blur_image(Z, 3)
    plt.figure()
    plt.imshow(Z)
    plt.figure()
    plt.imshow(Z2)
    plt.show()
    



#
# This file was generated using xslt from its XML file
#
# Copyright 2009, Associated Universities Inc., Washington DC
#
import sys
import os
from  casac import *
import string
from taskinit import casalog
from taskinit import xmlpath
#from taskmanager import tm
import task_concateovsa
def concateovsa(vis='', concatvis='', datacolumn='corrected', keep_orig_ms=True, cols2rm='model,corrected', freqtol='', dirtol='', respectname=False, timesort=True, copypointing=True, visweightscale=[], forcesingleephemfield=''):

        """Concatenate several EOVSA visibility data sets.
        concateovsa(vis=['UDB20180102161402.ms','UDB20180102173518.ms'], concatvis='UDB20180102_allday.ms')
        will concatenate 'UDB20180102161402.ms' and 'UDB20180102173518.ms' into 'UDB20180102_allday.ms'
    
        """
        if type(visweightscale)==float: visweightscale=[visweightscale]

#
#    The following is work around to avoid a bug with current python translation
#
        mytmp = {}

        mytmp['vis'] = vis
        mytmp['concatvis'] = concatvis
        mytmp['datacolumn'] = datacolumn
        mytmp['keep_orig_ms'] = keep_orig_ms
        mytmp['cols2rm'] = cols2rm
        mytmp['freqtol'] = freqtol
        mytmp['dirtol'] = dirtol
        mytmp['respectname'] = respectname
        mytmp['timesort'] = timesort
        mytmp['copypointing'] = copypointing
        mytmp['visweightscale'] = visweightscale
        mytmp['forcesingleephemfield'] = forcesingleephemfield
	pathname="file:///Users/fisher/Dropbox/PycharmProjects/suncasa/tasks/"
	trec = casac.utils().torecord(pathname+'concateovsa.xml')

        casalog.origin('concateovsa')
        if trec.has_key('concateovsa') and casac.utils().verify(mytmp, trec['concateovsa']) :
	    result = task_concateovsa.concateovsa(vis, concatvis, datacolumn, keep_orig_ms, cols2rm, freqtol, dirtol, respectname, timesort, copypointing, visweightscale, forcesingleephemfield)

	else :
	  result = False
        return result



import numpy as np
import os, platform
import astropy.units as u
from astropy import constants as const
import ctypes
from numpy.ctypeslib import ndpointer
from scipy import interpolate
import warnings

warnings.simplefilter("default")


def initGET_MW(libname):
    """
    Python wrapper for fast gyrosynchrotron codes.
    Identical to GScodes.py in https://github.com/kuznetsov-radio/gyrosynchrotron
    This is for the single thread version
    @param libname: path for locating compiled shared library
    @return: An executable for calling the GS codes in single thread
    """
    _intp = ndpointer(dtype=ctypes.c_int32, flags='F')
    _doublep = ndpointer(dtype=ctypes.c_double, flags='F')

    libc_mw = ctypes.CDLL(libname)
    mwfunc = libc_mw.pyGET_MW
    mwfunc.argtypes = [_intp, _doublep, _doublep, _doublep, _doublep, _doublep, _doublep]
    mwfunc.restype = ctypes.c_int

    return mwfunc


def sfu2tb(frequency, flux, area=None, size=None, square=True, reverse=False, verbose=False):
    """
        frequency: single element or array, in Hz
        flux: single element or array of flux, in sfu; if reverse, it is brightness temperature in K
        area: area in arcsec^2
        size: Two-dimensional width of the radio source, [major, minor], in arcsec.
              Ignored if both area and size are provided
        reverse: if True, convert brightness temperature in K to flux in sfu integrated uniformly withing the size
    """

    c = const.c.cgs
    k_B = const.k_B.cgs
    sfu = u.jansky * 1e4

    if (not 'area' in vars()) and (not 'size' in vars()):
        print('Neither area nor size is provided. Abort...')

    if not hasattr(frequency, 'unit'):
        # assume frequency is in Hz
        frequency = frequency * u.Hz

    if not hasattr(flux, 'unit'):
        # assume flux is in sfu
        if reverse:
            flux = flux * u.K
        else:
            flux = flux * sfu

    has_area = area is not None
    has_size = size is not None
    if has_area:
        if not hasattr(area, 'unit'):
            # assume area is in arcsec^2
            area = area * u.arcsec ** 2

    if has_size and (not has_area):
        if not isinstance(size, list):
            size = [size]

        if len(size) > 2:
            print('size needs to have 1 or 2 elements.')
        elif len(size) < 2:
            if verbose:
                print('Only one element is provided for source size. Assume symmetric source')
            if not hasattr(size[0], 'unit'):
                # assume size in arcsec
                size[0] = size[0] * u.arcsec
            # define half size
            a = b = size[0] / 2.
        else:
            if not hasattr(size[0], 'unit'):
                # assume size in arcsec
                size[0] = size[0] * u.arcsec
            if not hasattr(size[1], 'unit'):
                # assume size in arcsec
                size[1] = size[1] * u.arcsec
            # define half size
            a = size[0] / 2.
            b = size[1] / 2.
        if square:
            if verbose:
                print('Assume square-shaped source.')
            area = 4. * a * b
        else:
            if verbose:
                print('Assume elliptical-shaped source.')
            area = np.pi * a * b

    sr = area.to(u.radian ** 2)
    factor = c ** 2. / (2. * k_B * frequency ** 2. * sr)

    if reverse:
        # returned value is flux in sfu
        if verbose:
            print('converting input brightness temperature in K to flux density in sfu.')
        return (flux / factor).to(sfu, equivalencies=u.dimensionless_angles())
    else:
        # returned value is brightness temperature in K
        if verbose:
            print('converting input flux density in sfu to brightness temperature in K.')
        return (flux * factor).to(u.K, equivalencies=u.dimensionless_angles())


def ff_emission(em, T=1.e7, Z=1., mu=1.e10):
    T = T * u.k
    mu = mu * u.Hz
    esu = const.e.esu
    k_B = const.k_B.cgs
    m_e = const.m_e.cgs
    c = const.c.cgs
    bmax = (3 * k_B * T * u.k / m_e) ** 0.5 / 2.0 / np.pi / (mu * u.Hz)
    bmin = Z * esu ** 2 / 3. / k_B / T
    lnbb = np.log((bmax / bmin).value)
    ka_mu = 1. / mu ** 2 / T ** 1.5 * (
            Z ** 2 * esu ** 6 / c / np.sqrt(2. * np.pi * (m_e * k_B) ** 3)) * np.pi ** 2 / 4.0 * lnbb

    opc = ka_mu * em
    return T.value * (1 - np.exp(-opc.value))


class GSCostFunctions:
    def SinglePowerLawMinimizerOneSrc(fit_params, freqghz, spec=None, spec_err=None,
                                      spec_in_tb=True, pgplot_widget=None, show_plot=False, debug=False, verbose=False):
        """
        params: parameters defined by lmfit.Paramters()
        freqghz: frequencies in GHz
        spec: input spectrum, can be brightness temperature in K, or flux density in sfu
        spec_err: uncertainties of spectrum in K or sfu
        spec_in_tb: if True, input is brightness temperature in K, otherwise is flux density in sfu
        calc_flux: Default (False) is to return brightness temperature.
                    True if return the calculated flux density. Note one needs to provide src_area/src_size for this
                        option. Otherwise assumes src_size = 2 arcsec (typical EOVSA pixel size).
        @rtype: 1. If no tb/tb_err or flux/flux_err is provided, return the calculated
                    brightness temperature or flux for each input frequency.
                2. If tb/tb_err or flux/flux_err are provided, return the
                    (scaled) residual for each input frequency
        """
        if platform.system() == 'Linux' or platform.system() == 'Darwin':
            libname = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   'binaries/MWTransferArr.so')
        if platform.system() == 'Windows':  ##TODO: not yet tested on Windows platform
            libname = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   'binaries/MWTransferArr64.dll')
        GET_MW = initGET_MW(libname)  # load the library

        asec2cm = 0.725e8
        if 'area_asec2' in fit_params.keys():
            src_area = float(fit_params['area_asec2'].value)  # source area in arcsec^2
        else:
            src_area = 4.  # arcsec^2. default area for bright temperature spectral fitting. Will be divided out.
            if not spec_in_tb:
                print('=======Warning: no source area is provided for flux density calculation. '
                      'Use area = 4 arcsec^2 as the place default (1 EOVSA pixel).======')
        src_area_cm2 = src_area * asec2cm ** 2.  # source area in cm^2
        depth_cm = float(fit_params['depth_asec'].value) * asec2cm  # total source depth in cm
        Bmag = float(fit_params['Bx100G'].value) * 100.  # magnetic field strength in G
        Tth = float(fit_params['T_MK'].value) * 1e6  # thermal temperature in K
        nth = 10. ** float(fit_params['log_nth'].value)  # thermal density
        nrl = 10. ** float(fit_params['log_nnth'].value)  # total nonthermal density above E_min
        delta = float(fit_params['delta'].value)  # powerlaw index
        theta = float(fit_params['theta'].value)  # viewing angle in degrees
        Emin = float(fit_params['Emin_keV'].value) / 1e3  # low energy cutoff of nonthermal electrons in MeV
        Emax = float(fit_params['Emax_MeV'].value)  # high energy cutoff of nonthermal electrons in MeV
        if debug:
            # debug against previous codes
            print('depth, Bmag, Tth, nth/1e10, lognrl, delta, theta, Emin, Emax: '
                  '{0:.1f}, {1:.1f}, {2:.1f}, {3:.1f}, {4:.1f}, '
                  '{5:.1f}, {6:.1f}, {7:.2f}, {8:.1f}'.format(depth_cm / 0.725e8, Bmag, Tth / 1e6, nth / 1e10,
                                                              np.log10(nrl), delta, theta, Emin, Emax))
        # E_hi = 0.1
        # nrl = nrlh * (Emin ** (1. - delta) - Emax * (1. - delta)) / (E_hi ** (1. - delta) - Emax ** (1. - delta))

        Nf = 100  # number of frequencies
        NSteps = 1  # number of nodes along the line-of-sight

        Lparms = np.zeros(11, dtype='int32')  # array of dimensions etc.
        Lparms[0] = NSteps
        Lparms[1] = Nf

        Rparms = np.zeros(5, dtype='double')  # array of global floating-point parameters
        Rparms[0] = src_area_cm2  # Area, cm^2
        Rparms[1] = 0.8e9  # starting frequency to calculate spectrum, Hz
        Rparms[2] = 0.02  # logarithmic step in frequency
        Rparms[3] = 0  # f^C
        Rparms[4] = 0  # f^WH

        ParmLocal = np.zeros(24, dtype='double')  # array of voxel parameters - for a single voxel
        ParmLocal[0] = depth_cm / NSteps  # voxel depth, cm
        ParmLocal[1] = Tth  # T_0, K
        ParmLocal[2] = nth  # n_0 - thermal electron density, cm^{-3}
        ParmLocal[3] = Bmag  # B - magnetic field, G
        ParmLocal[6] = 3  # distribution over energy (PLW is chosen)
        ParmLocal[7] = nrl  # n_b - nonthermal electron density, cm^{-3}
        ParmLocal[9] = Emin  # E_min, MeV
        ParmLocal[10] = Emax  # E_max, MeV
        ParmLocal[12] = delta  # \delta_1
        ParmLocal[14] = 0  # distribution over pitch-angle (isotropic is chosen)
        ParmLocal[15] = 90  # loss-cone boundary, degrees
        ParmLocal[16] = 0.2  # \Delta\mu

        Parms = np.zeros((24, NSteps), dtype='double', order='F')  # 2D array of input parameters - for multiple voxels
        for i in range(NSteps):
            Parms[:, i] = ParmLocal  # most of the parameters are the same in all voxels
            Parms[4, i] = theta

        RL = np.zeros((7, Nf), dtype='double', order='F')  # input/output array
        dummy = np.array(0, dtype='double')

        # calculating the emission for array distribution (array -> on)
        res = GET_MW(Lparms, Rparms, Parms, dummy, dummy, dummy, RL)
        if res:
            # retrieving the results
            f = RL[0]
            I_L = RL[5]
            I_R = RL[6]
            all_zeros = not RL.any()
            if not all_zeros:
                flux_model = I_L + I_R
                flux_model = np.nan_to_num(flux_model) + 1e-11
                logf = np.log10(f)
                logflux_model = np.log10(flux_model)
                logfreqghz = np.log10(freqghz)
                interpfunc = interpolate.interp1d(logf, logflux_model, kind='linear')
                logmflux = interpfunc(logfreqghz)
                mflux = 10. ** logmflux
                mtb = sfu2tb(np.array(freqghz) * 1.e9, mflux, area=src_area).value

                if pgplot_widget:
                    ##todo: figure out a way to update the main widget
                    import pyqtgraph as pg
                    all_items = pgplot_widget.getPlotItem().listDataItems()
                    if len(all_items) > 0:
                        pgplot_widget.removeItem(all_items[-1])
                    spec_fitplot = pgplot_widget.plot(x=np.log10(freqghz), y=np.log10(mtb),
                                                      pen=dict(color=pg.mkColor(0), width=3),
                                                      symbol=None, symbolBrush=None)
                    pgplot_widget.addItem(spec_fitplot)

                if show_plot:
                    import matplotlib.pyplot as plt
                    fig, (ax1, ax2) = plt.subplots(1, 2)
                    ax1.plot(freqghz, mflux, 'k')
                    # ax1.set_xlim([1, 20])
                    ax1.set_xlabel('Frequency (GHz)')
                    ax1.set_ylabel('Flux (sfu)')
                    ax1.set_title('Flux Spectrum')
                    ax1.set_xscale('log')
                    ax1.set_yscale('log')
                    ax2.plot(freqghz, mtb, 'k')
                    # ax2.set_xlim([1, 20])
                    ax2.set_xlabel('Frequency (GHz)')
                    ax2.set_ylabel('Brightness Temperature (K)')
                    ax2.set_title('Brightness Temperature Spectrum')
                    ax2.set_xscale('log')
                    ax2.set_yscale('log')
                    ax2.legend()
                    plt.show()
            else:
                print("Calculation error! Assign an unrealistically huge number")
                mflux = np.ones_like(freqghz) * 1e4
                mtb = sfu2tb(np.array(freqghz) * 1.e9, mflux, area=src_area).value
        else:
            print("Calculation error! Assign an unrealistically huge number")
            mflux = np.ones_like(freqghz) * 1e9
            mtb = sfu2tb(np.array(freqghz) * 1.e9, mflux, area=src_area).value

        # Return values
        if spec_in_tb:
            if spec is None:
                # nothing is provided, return the model spectrum
                return mtb
            if spec_err is None:
                # no uncertainty provided, return absolute residual
                return mtb - spec
            # Return scaled residual
            return (mtb - spec) / spec_err
        else:
            if spec is None:
                # nothing is provided, return the model spectrum
                return mflux
            if spec_err is None:
                # no uncertainty provided, return absolute residual
                return mflux - spec
            # Return scaled residual
            return (mflux - spec) / spec_err



#!/usr/bin/env python
import numpy as np
import numpy.ma as ma
import copy, struct
import shutil
import os
from math import *
import dspec 
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
from fnmatch import fnmatch
import pdb
from taskinit import *
import optparse
import pyfits
from glob import glob

stokesdict={'I':1,'Q':2,'U':3,'V':4,'RR':-1,'LL':-2,'RL':-3,'LR':-4,'XX':-5,'YY':-6,'XY':-7,'YX':-8}

def get_vspec(fitspath=None,vspecfile=None,tinc=1,rgns=None):
    #setup initial parametesr
    #fitspath = '/home/bchen/work/EVLA/20120303/S16-20/fits/rgn1/'
    #rgns=[[[95,112],[120,140]],
    #      [[129,109],[163,141]],
    #      [[111,66],[135,98]]]
    #      [[182,135],[213,168]]]
    # check if the path exist
    if not os.path.exists(fitspath):
        print 'specified path '+fitspath+' does not exist. aborting...'
        return
    fitsfiles_= glob(fitspath+'/*.fits')
    if len(fitsfiles_) < 1:
        print 'no fits files found. aborting...'
        return
    fitsfiles_.sort()
    fitsfiles=fitsfiles_[::tinc]
    ntim=len(fitsfiles)
    tims=np.zeros(ntim)
    timstrs=[]

    #determine the frequency axis
    fitsfile0=fitsfiles[0]
    ia.open(fitsfile0)
    info0=ia.summary()
    dim=info0['shape']
    nx0=dim[0]
    ny0=dim[1]
    nf0=dim[2]
    npol0=dim[3]
    stokes=[]
    svals=(np.arange(npol0)*info0['incr'][3]+info0['refval'][3])
    for sval in svals:
        stokes.append(stokesdict.keys()[stokesdict.values().index(sval)])
    ia.close()
          
    nrgn=len(rgns)
    fmaxs=np.zeros((nf0,ntim,npol0,nrgn))
    fsums=np.zeros((nf0,ntim,npol0,nrgn))
    fmeans=np.zeros((nf0,ntim,npol0,nrgn))
    print str(ntim)+' times to process...'
    timstrs=[]
    tims = np.zeros(ntim)
    for t in range(ntim):
        fitsfile=fitsfiles[t]
        #get fits information 
        hdu=pyfits.open(fitsfile)
        hdr = hdu[0].header
        timstr = hdr['date-obs']
        timstrs.append(timstr)
        tims[t] = qa.convert(timstr,'s')['value']
        exptime=hdr['exptime']
        ndim = hdr['naxis']
        if ndim != 4:
            print 'input dimension is not 4. skipping this fits file'
            break
        nx = hdr['naxis1']
        ny = hdr['naxis2']
        nf = hdr['naxis3']
        npol = hdr['naxis4']
        crpixs = [hdr['crpix1'],hdr['crpix2'],hdr['crpix3'],hdr['crpix4']]
        crvals = [hdr['crval1'],hdr['crval2'],hdr['crval3'],hdr['crval4']]
        cdelts = [hdr['cdelt1'],hdr['cdelt2'],hdr['cdelt3'],hdr['cdelt4']]
        freqs=(np.arange(nf)*cdelts[2]+crvals[2])/1e6
        hdu.close()

        if t % 50 == 0 and t > 0:
            print 'Processed '+str(t)+' times'
            print 'Current time: ', timstr
        ia.open(fitsfile)
        for p in range(npol):
            for k in range(nrgn):
                rgn=rgns[k]
                r1=rg.box(blc=rgn[0],trc=rgn[1])
                try:
                    mystat=ia.statistics(axes=[0,1],region=r1)
                    fmax=mystat['max']
                    fsum=mystat['sum']
                    fmean=mystat['mean']
                    npt=mystat['npts']
                    fmaxs[:,t,p,k]=fmax
                    fsums[:,t,p,k]=fsum
                    fmeans[:,t,p,k]=fmean
                except:
                    print 'failure for this time: ',timstrs[t]
        ia.close()
    if not vspecfile:
        vspecfile='./vspec.t'+timstrs[0].replace(':','')+'-'+\
                  timstrs[-1].replace(':','')+'.npz'
    np.savez(vspecfile,rgns=rgns,freqs=freqs,tims=tims,timstrs=timstrs,
             stokes=stokes,fmaxs=fmaxs,fsums=fsums,fmeans=fmeans)

def wrt_vspec(vspecfile=None,vspecdat=None):
    try:
        vspecfile
    except NameError:
        print 'No input centfile specified for reading. Aborting...'
    if not vspecdat:
        print 'Output file name is not specified, use the default convention'
        vspecdat=vspecfile.replace('npz','dat')
    vspecdata=np.load(vspecfile)
    rgns=vspecdata['rgns']
    nrgn=rgns.shape[0]
    stokes=vspecdata['stokes']
    nstokes=len(stokes)
    fmx=copy.deepcopy(vspecdata['fmaxs'][:,:,:,:])
    fmean=copy.deepcopy(vspecdata['fmeans'][:,:,:,:])
    nf,nt,npl,nr=fmx.shape
    print 'Dimension of the data cube -- # of freq, # of time, # of pol, # of regions:'
    print nf, nt, npl, nr
    # Turn fmx into a 1d array for writing
    fmx.shape = (nf*nt*npl*nr)
    fmean.shape = (nf*nt*npl*nr)
    # Pack the data into a byte buffer.  Note the asterisks in the next three lines!
    # If your data includes integers or other data types, use 'i' or other instead of 'f'
    buf = struct.pack(str(nf)+'f',*vspecdata['freqs'])
    buf += struct.pack(str(nt)+'f',*vspecdata['tims'])
    buf += struct.pack(str(nf*nt*npl*nr)+'f',*fmx)
    buf += struct.pack(str(nf*nt*npl*nr)+'f',*fmean)
    with open(vspecdat,'wb') as f:
        f.write(buf)
    f.close()

def plt_vspec(vspecfile=None,imgfile=None,moment=None, pltmax=True, pltmean=False, bkgsub=False, \
              dmin=None,dmax=None,savepng=True,savepdf=False,\
              timerange=None, freqrange=None,timestr=False):
    # vspecfile: saved npz file containing the vector dynamic spectrum data (from get_max)
    # imagefile: image files to overplot selected regions, the number of images 
    #               must equal to len(stokes)
    # pltmax: plot the maximum value of each selected region
    # pltmin: plot the mean value of each selected region
    # bkgsub: if True, subtract max/mean of the first region from all the other regions
    # dmin: minimum flux to show, one- or len(stokes) element list 
    # dmax: maximum flux to show, one- or len(stokes) element list
    # savepng: if True, save the resulted figure as a png file
    # savepdf: if True, save the resulted figure as a pdf file
    # timerange: format: ['18:00:00','19:00:00']
    # freqrange: format: [1000.,1500.]
    try:
        vspecfile
    except NameError:
        print 'No input centfile specified for reading. Aborting...'
    vspecdata=np.load(vspecfile)
    rgns=vspecdata['rgns']
    nrgn=rgns.shape[0]
    fmaxs=vspecdata['fmaxs']
    fmeans=vspecdata['fmeans']
    tims=vspecdata['tims']
    ntim=tims.shape
    freqs=vspecdata['freqs']
    nfreq=freqs.shape
    stokes=vspecdata['stokes']
    nstokes=len(stokes)
    if pltmax:
        fluxs=fmaxs
    if pltmean:
        fluxs=fmeans
    if not moment:
        moment = 0
    if bkgsub:
        if nrgn < 2:
            print 'Background subtraction requires more than one regions'
            return
        for r in range(nrgn-1):
            fluxs[:,:,:,r+1]-=fluxs[:,:,:,0] #the first region is the background
    if (not dmin) or (not dmax):
        print 'will use automatic brightness scaling'
    elif len(dmin) == 1 and len(dmax) == 1:
        dmin=nstokes*dmin
        dmax=nstokes*dmax
    elif len(dmin) != nstokes or len(dmax) != nstokes:
        raise ValueError('Number of dmin or dmax does not agree with number of stokes')
    if timerange:
        if type(timerange[0]) is str:
            timerange=[qa.convert(qa.quantity(t),'s')['value'] for t in timerange]
        tidx=np.where((tims >= timerange[0]) & (tims <= timerange[1]))[0]
    else:
        tidx=range(ntim[0])
    if freqrange:
        fidx=np.where((freqs >= freqrange[0]) & (freqs <= freqrange[1]))[0]
    else:
        fidx=range(nfreq[0])
    
    #set up time axis
    #1 sec per tick
    tstart=np.fix(tims[0])
    xticks=np.arange(ntim[0]/20)+tstart
    nticks=xticks.shape
    xticknames=[]
    for i in range(nticks[0]):
        tim=xticks[i]
        #timstr=qa.time(qa.quantity(tim,'s'),prec=6)[0]
        timstr="%d" % (tim-tstart)
        xticknames.append(timstr)
    xticks=list(xticks)

    #do the plot
    #plot moment image
    if imgfile:
        try:
            ia.open(imgfile)
        except NameError:
            print 'Specified image file can not be opened'
        f,ax = plt.subplots(figsize=(5,5))
        im2=ia.moments(moments=[moment],excludepix=[-1000,0])
        pix=im2.getchunk()
        im2.done()
        ia.close()
        imdim=pix.shape
        immom=np.transpose(pix[:,:,0])
        ax.imshow(immom,aspect='auto',origin='lower',extent=[0,imdim[0]-1,0,imdim[1]-1])
        ax.set_autoscale_on(False)
        ax.text(0.03,0.93,'Momment '+str(moment)+' image',color='w')
        for rgn in rgns:
            xbg=rgn[0][0]
            xend=rgn[1][0]
            ybg=rgn[0][1]
            yend=rgn[1][1]
            ax.axvspan(xbg,xend,ymin=float(ybg)/float(imdim[1]),ymax=float(yend)/float(imdim[1]),
                        alpha=1.0,fill=False,lw=1,color='w')
    figsize=(8,10)
    nrgns=rgns.shape[0]
    if bkgsub:
        nplot=nrgns-1
    else:
        nplot=nrgns
    f2,axarr = plt.subplots(nplot,nstokes,figsize=figsize)
    for p in range(nstokes):
        for n in range(nplot):
            if nplot == 1 and nstokes == 1:
                ax=axarr
            if nplot > 1 and nstokes == 1:
                ax=axarr[n]
            if nplot > 1 and nstokes > 1:
                ax=axarr[n,p]
            if bkgsub:
                flux=fluxs[:,:,p,n+1]
            else:
                flux=fluxs[:,:,p,n]
            mflux=ma.masked_outside(flux,10,1e6)
            flux_med=ma.median(mflux)
            if not dmin or not dmax:
                vmin=flux_med/3.
                vmax=flux_med*8.
            else:
                vmin=dmin[p]
                vmax=dmax[p]
            ax.pcolormesh(tims,freqs,flux,vmin=vmin,vmax=vmax)
            ax.set_xlim(tims[tidx[0]],tims[tidx[-1]])
            ax.set_ylim(freqs[fidx[0]],freqs[fidx[-1]])
            ax.set_ylabel('Frequency (MHz)')
            if timestr:
                labels=ax.get_xticks().tolist()
                newlabels=[qa.time(qa.quantity(lb,'s'))[0] for lb in labels] 
                ax.set_xticklabels(newlabels)
            #ax.text(0.05,0.04,'1',color='k',transform=ax.transAxes)
    if savepng:
        figfile=vspecfile[:(vspecfile.find('npz'))]+'png'
        f2.savefig(figfile)
    if savepdf:
        figfile=vspecfile[:(vspecfile.find('npz'))]+'pdf'
        f2.savefig(figfile)

def plt_max(vspecfile='',fitsfile='',pltmax=True, pltmean=False, bkgsub=True,rgns='',dmin='',dmax='',figfile=''):
    #vspecfile = vspecpath+'maxfit.t183100.025-183109.974.I.npz'
    #fitspath = '/lustre/bchen/CASA/MAR3/S16-20/fits/rgn1/'
    #fitsfile = 'SUN_66660.025.fits'
    vspecdata=np.load(vspecfile)
    rgns=vspecdata['rgns']
    nrgn=rgns.shape[0]
    fmaxs=vspecdata['fmaxs']
    fmeans=vspecdata['fmeans']
    tims=vspecdata['tims']
    freqs=vspecdata['freqs']
    if pltmax:
        fluxs=fmaxs
    if pltmean:
        fluxs=fmeans
    if bkgsub:
        if nrgn < 2:
            print 'Background subtraction requires more than one regions'
            return
        for r in range(nrgn-1):
            fluxs[:,:,:,r+1]-=fluxs[:,:,:,0] #the first region is the background
    
    #setup plot parameters
    fluxs_med=np.median(fluxs)
    if not dmin:
        dmin=fluxs_med/20.
    if not dmax:
        dmax=fluxs_med*2.
    #fill missing times with zeros
    #find out the time range
    timeran=tims[-1]-tims[0]
    #find out the integration interval
    ntim=len(tims)
    t_ints=[]
    iter=0
    for i in range(ntim-1):
        t_ints.append(tims[i+1]-tims[i])
    t_int=np.median(t_ints)
    ntim_plt=round(timeran/t_int)

    xi=np.linspace(tims[0],tims[-1],900)
    yi=np.linspace(freqs[0],freqs[-1],128)
    fluxs_plt=griddata((tims,freqs),fluxs,(xi,yi),method='cubic')
    #1 sec per tick
    tstart=np.fix(tims[0])
    xticks=np.arange(ntim[0]/20)+tstart
    nticks=xticks.shape
    xticknames=[]
    for i in range(nticks[0]):
        tim=xticks[i]
        #timstr=qa.time(qa.quantity(tim,'s'),prec=6)[0]
        timstr="%d" % (tim-tstart)
        xticknames.append(timstr)
    xticks=list(xticks)

    #do the plot
    figsize=(7.36,8)
    f=plt.figure(figsize=figsize,dpi=100)
    #rect1=[0.12,0.52,0.27,0.46]
    #rect2=[0.45,0.52,0.50,0.46]
    #rect3=[0.12,0.06,0.27,0.42]
    #rect4=[0.40,0.06,0.27,0.42]
    #rect5=[0.68,0.06,0.27,0.42]
    rect1=[0.12,0.52,0.27,0.46]
    rect2=[0.45,0.52,0.50,0.46]
    rect3=[0.12,0.06,0.20,0.42]
    rect4=[0.33,0.06,0.20,0.42]
    rect5=[0.54,0.06,0.20,0.42]
    rect6=[0.75,0.06,0.20,0.42]
    ax1=f.add_axes(rect1)
    ax2=f.add_axes(rect2)
    ax3=f.add_axes(rect3)
    ax4=f.add_axes(rect4)
    ax5=f.add_axes(rect5)
    ax6=f.add_axes(rect6)
    plt.figtext(0.3,0.01,'Time in seconds, starting from 18:30:02')
    #plot cross-power dspec
    specpath='/lustre/bchen/CASA/MAR3/S16-20/'
    specfile=specpath+'sun.mar3.S16-20.cal.grd.ms.bl19-22.spec.npz'
    specdata=np.load(specfile)
    spec=specdata['spec']
    xtim=specdata['tim']
    xtimsecs=np.zeros(xtim.shape)
    #convert to seconds from 2012-03-03 00:00
    for i in range(xtim.shape[0]):
        xtimsecs[i]=qa.sub(qa.quantity(xtim[i],'s'),qa.quantity('2012-03-03','s'))['value']
    xfreq=specdata['freq']
    tbgind=np.where(xtimsecs>tims[0])[0][0]
    tendind=np.where(xtimsecs<tims[-1])[0][-1]
    fbgind=np.where(xfreq/1e6>=freqs[0])[0][0]
    fendind=np.where(xfreq/1e6<=freqs[-1])[0][-1]
    specsel=np.absolute(spec[:,fbgind:fendind+1,tbgind:tendind+1])
    #pdb.set_trace()
    ax1.imshow((specsel[0,:,:]+specsel[1,:,:])/2,aspect='auto',origin='lower',extent=[tims[0],tims[-1],freqs[0],freqs[-1]])
    ax1.set_autoscale_on(False)
    ax1.set_xticks(xticks)
    ax1.set_xticklabels(xticknames)
    ax1.set_ylabel('Frequency (MHz)')
    ax1.text(0.05,0.93,'x-power spec',color='k',transform=ax1.transAxes)
    #plot moment image
    ia.open(fitspath_+fitsfile)
    im2=ia.moments(moments=[0],excludepix=[-1000,0])
    pix=im2.getchunk()
    im2.done()
    ia.close()
    imdim=pix.shape
    immom=np.transpose(pix[:,:,0])
    ax2.imshow(immom,aspect='auto',origin='lower',extent=[0,imdim[0]-1,0,imdim[1]-1])
    ax2.set_autoscale_on(False)
    ax2.text(100,150,'1',color='w')
    ax2.text(150,150,'2',color='w')
    ax2.text(120,50,'3',color='w')
    ax2.text(170,160,'4',color='w')
    ax2.text(0.03,0.93,'Image at 18:31:00 (moment '+str(moment)+')',color='w',transform=ax2.transAxes)
    #rgns=[[[95,112,0,0],[120,140,0,0]],
    #      [[129,109,0,0],[163,141,0,0]],
    #      [[111,66,0,0],[135,98,0,0]]]
    for rgn in rgns:
        xbg=rgn[0][0]
        xend=rgn[1][0]
        ybg=rgn[0][1]
        yend=rgn[1][1]
        ax2.axvspan(xbg,xend,ymin=float(ybg)/float(imdim[1]),ymax=float(yend)/float(imdim[1]),
                    alpha=1.0,fill=False,lw=1,color='w')
    ax3.imshow(fluxs[:,:,0],aspect='auto',origin='lower',extent=[tims[0],tims[-1],freqs[0],freqs[-1]])
    ax3.set_xticks(xticks)
    ax3.set_xticklabels(xticknames)
    #ax2.set_xlabel('Universal Time (50 ms/pixel)')
    ax3.set_ylabel('Frequency (MHz)')
    #ax2.set_title('region 3')
    ax3.text(0.05,0.04,'1',color='k',transform=ax3.transAxes)
    ax4.imshow(fluxs[:,:,1],aspect='auto',origin='lower',extent=[tims[0],tims[-1],freqs[0],freqs[-1]])
    ax4.set_xticks(xticks)
    ax4.set_xticklabels(xticknames)
    #ax3.set_xlabel('Universal Time (50 ms/pixel)')
    ax4.set_yticks([])
    ax4.set_ylabel('')
    #ax3.set_title('region 3')
    ax4.text(0.05,0.04,'2',color='w',transform=ax4.transAxes)
    ax5.imshow(fluxs[:,:,2],aspect='auto',origin='lower',extent=[tims[0],tims[-1],freqs[0],freqs[-1]])
    ax5.set_xticks(xticks)
    ax5.set_xticklabels(xticknames)
    #ax4.set_xlabel('Universal Time (50 ms/pixel)')
    ax5.set_yticks([])
    ax5.set_ylabel('')
    #ax4.set_title('region 3')
    ax5.text(0.05,0.04,'3',color='k',transform=ax5.transAxes)
    ax6.imshow(fluxs[:,:,3],aspect='auto',origin='lower',extent=[tims[0],tims[-1],freqs[0],freqs[-1]])
    ax6.set_xticks(xticks)
    ax6.set_xticklabels(xticknames)
    #ax4.set_xlabel('Universal Time (50 ms/pixel)')
    ax6.set_yticks([])
    ax6.set_ylabel('')
    #ax4.set_title('region 3')
    ax6.text(0.05,0.04,'4',color='k',transform=ax6.transAxes)
    if not figfile:
        figfile=centfile[:(centfile.find('npz'))]+'pdf'
    else:
        figfile=figfile
    f.savefig(figfile,dpi=300)

if __name__ == "__main__":
    import sys
    get_max(sys.argv[1],sys.argv[2])
